import logging as lg
import os
import os.path as osp

from six.moves.configparser import ConfigParser

_logger = lg.getLogger(__name__)


class Config(object):
    config_file = "tanit.conf"
    default_conf_dir = osp.expanduser("/etc/tanit/conf")

    def __init__(self, path=None):
        self.config = ConfigParser()

        self.conf_dir = path or os.getenv("TANIT_CONF_DIR", self.default_conf_dir)
        self.conf_file = osp.join(self.conf_dir, self.config_file)

        if osp.exists(self.conf_file):
            try:
                self.config.read(self.conf_file)
                try:
                    self.load()
                except TanitConfigurationException as e:
                    _logger.error(
                        "Error loading configuration from file %s.", self.conf_file
                    )
                    raise e

            except Exception as e:
                raise TanitConfigurationException(
                    "Exception while loading configuration file %s.", self.conf_file, e
                )

            _logger.info("Instantiated configuration from %s.", self.conf_file)
        else:
            raise TanitConfigurationException(
                "Invalid configuration file %s.", self.conf_file
            )


class TanitConfigurationException(Exception):
    pass



import logging as lg

from ...common.model.worker import Worker
from ...filesystem.filesystem_factory import FileSystemFactory
from .execution.execution_manager import ExecutionManager
from .worker.worker_decommissioner import WorkerDecommissioner
from .worker.worker_factory import WorkerFactory
from .worker.worker_manager import WorkerManager

_logger = lg.getLogger(__name__)


class Master(object):
    """
    The master object represent the service interface of the master server.

    Its role is to implement the master API in a technology independent way,
    so the service layer is completely dissociated from the service layer
    and the RPC technology used.
    """

    def __init__(self, config=None):
        # filesystems factory
        self.filesystems_factory = FileSystemFactory.getInstance()
        # workers factory
        self.workers_factory = WorkerFactory(self)
        # workers manager
        self.workers_manager = WorkerManager(self.workers_factory, config)
        # execution manager
        self.execution_manager = ExecutionManager(self.workers_manager, config)
        # decommissioner
        self.decommissioner = WorkerDecommissioner(
            self.execution_manager, self.workers_manager
        )

        self.configure(config)

        self.started = False

    def configure(self, config):
        pass

    def submit_job(self, job):
        if not self.started:
            raise MasterStoppedException("Can not submit job, master server stopped.")
        return self.execution_manager.submit_job(job)

    def list_jobs(self):
        return self.execution_manager.list_jobs()

    def get_job(self, jid):
        return self.execution_manager.get_job(jid)

    def task_start(self, tid):
        self.execution_manager.task_start(tid)

    def task_success(self, tid):
        self.execution_manager.task_finish(tid)

    def task_failure(self, tid):
        self.execution_manager.task_failure(tid)

    def list_workers(self):
        _logger.info("Listing Workers.")
        wkr_list = []
        for wkr in self.workers_manager.list_live_workers():
            wkr_list.append(Worker(wkr.wid, wkr.address, wkr.port))
        return wkr_list

    def register_worker(self, worker):
        if not self.started:
            raise MasterStoppedException(
                "Can not register worker [ %s ] : master server stopped.", worker.wid
            )

        _logger.info("Registering new Worker [ %s ].", worker.wid)
        # register the worker as an executor in the workers manager
        self.workers_manager.register_worker(worker)
        _logger.info("Worker [ %s ] registered.", worker.wid)

    def register_filesystem(self, name, filesystem):
        if not self.started:
            raise MasterStoppedException(
                "Can not register filesystem [ %s ] : master server stopped.", name
            )

        _logger.info("Registering new filesystem [ %s ].", name)
        filesystem["name"] = name
        # register the worker as a filesystem
        self.filesystems_factory.register_filesystem(filesystem)
        # notify the workers about the new file system
        for worker in self.workers_manager.list_live_workers():
            worker.register_filesystem(name, filesystem)
        _logger.info("Filesystem [ %s ] registered.", name)

    def register_heartbeat(self, worker):
        _logger.debug("Received heart beat from Worker [ %s ].", worker.wid)
        self.workers_manager.register_heartbeat(worker)

    def unregister_worker(self, worker):
        if not self.started:
            raise MasterStoppedException(
                "Can not register worker [ %s ] : master server stopped.", worker.wid
            )

        # This will prevent any future tasks from being sent to the worker
        self.workers_manager.decommission_worker(worker.wid)

    def start(self):
        _logger.info("Stating Tanit master services.")
        self.started = True
        self.workers_manager.start()
        self.execution_manager.start()
        self.decommissioner.start()
        _logger.info("Tanit master services started.")

    def stop(self):
        _logger.info("Stopping Tanit master services.")
        self.started = False
        self.decommissioner.stop()
        self.decommissioner.join()
        self.execution_manager.stop()
        self.workers_manager.stop()
        _logger.info("Tanit master services stopped.")


class MasterStoppedException(Exception):
    """Raised when trying to submit a task to a stopped master."""

    pass



"""
tanit-client: A simple thrift client for tanit service.

Usage:
  tanit-client job [-v...] [--submit job-spec] [--status jid] [--list]
  tanit-client worker [-v...] [--list]
  tanit-client -h | --help
  tanit-client --version

Options:
  --version                     Show version and exit.
  -h --help                     Show help and exit.
  -v --verbose                  Enable log output. Can be specified up to three
                                times (increasing verbosity each time).
  --status JOB_ID               Print the status of a job.
  --submit job-spec             Submit a job to the tanit master.
  --list                        List all jobs.

Examples:
  tanit-client --submit "{
        "src": "prod",
        "dest": "dr",
        "src_path":
        "/data/important_dataset",
        "dest_path": "/data/important_dataset",
        "preserve": "False" }"

"""
import json
import logging as lg

from docopt import docopt

from ... import __version__
from ...common.model.execution_type import ExecutionType
from ...common.model.job import Job
from ..config.config import MasterConfig
from .client import ClientType
from .client import ThriftClientFactory

_logger = lg.getLogger(__name__)


def configure_logging():
    # capture warnings issued by the warnings module
    lg.captureWarnings(True)

    logger = lg.getLogger()
    logger.setLevel(lg.DEBUG)
    lg.getLogger("requests_kerberos.kerberos_").setLevel(lg.CRITICAL)
    lg.getLogger("requests").setLevel(lg.ERROR)

    # Configure stream logging if applicable
    stream_handler = lg.StreamHandler()
    stream_handler.setLevel(lg.INFO)

    fmt = "%(levelname)s\t%(message)s"
    stream_handler.setFormatter(lg.Formatter(fmt))
    logger.addHandler(stream_handler)


def main(argv=None):
    args = docopt(__doc__, argv=argv, version=__version__)

    configure_logging()
    config = MasterConfig()

    if args["job"]:
        client = ThriftClientFactory(
            config.client_service_host, config.client_service_port
        ).create_client(ClientType.USER_SERVICE)
        client.start()
        if args["--list"]:
            for job in client.list_jobs():
                print(str(job))
        elif args["--submit"]:
            try:
                if args["--submit"].startswith("@"):
                    with open(args["--submit"][1:], "r") as json_spec_file:  # NOQA
                        job_spec = json.load(json_spec_file)
                else:
                    job_spec = json.loads(args["--submit"])
            except Exception as e:
                _logger.error("Error parsing job json specification.")
                raise e

            jtype = ExecutionType._NAMES_TO_VALUES[job_spec["type"]]
            params = job_spec["params"]
            client.submit_job(Job(jtype, params))
        elif args["--status"]:
            job = client.job_status(args["--status"])
            if job is None:
                _logger.info("No such job %s", args["--status"])
            else:
                print(str(job))
        else:
            _logger.error("Nothing to do !")

        client.stop()
    elif args["worker"]:
        client = ThriftClientFactory(
            config.worker_service_host, config.worker_service_port
        ).create_client(ClientType.WORKER_SERVICE)
        client.start()
        if args["--list"]:
            for worker in client.list_workers():
                print(str(worker))
        else:
            _logger.error("Nothing to do !")

        client.stop()


if __name__ == "__main__":
    main()



import time

import pytest

from tanit.filesystem.filesystem_factory import FileSystemFactory

from .test_filesystem import BaseFilesystemTest


@pytest.fixture(scope="class")
def filesystem():
    filesystems_factory = FileSystemFactory.getInstance()
    filesystems_factory.register_filesystem(
        {"name": "gcs-kraken-test-bucket", "type": "gcs", "bucket": "kraken-test"}
    )
    filesystem = filesystems_factory.get_filesystem("gcs-kraken-test-bucket")
    yield filesystem


@pytest.fixture(scope="class")
def test_data(filesystem):
    test_dir = "/data/integration-test/dataset-%s" % int(time.time() * 1000)
    filesystem.mkdir(test_dir)
    yield test_dir
    filesystem.delete(test_dir, recursive=True)


class TestGCSFilesystem(BaseFilesystemTest):
    pass



from ...filesystem.ioutils import FileSystemError
from ...filesystem.local.local_filesystem import LocalFileSystem
from ...thrift.worker.filesystem.ttypes import FileContent
from ...thrift.worker.filesystem.ttypes import FileContentOrNull
from ...thrift.worker.filesystem.ttypes import FileStatus
from ...thrift.worker.filesystem.ttypes import FileStatusOrNull
from ...thrift.worker.filesystem.ttypes import FileSystemException


class LocalFileSystemHandler(object):
    def __init__(self):
        self.fs = LocalFileSystem()
        self.file_descs = {}

    def start(self):
        pass

    def stop(self):
        for file in self.file_descs.values():
            file.close()

    def ls(self, path, status, glob):
        try:
            _list = self.fs.list(path, status, glob)
        except Exception as e:
            raise FileSystemException(str(e))
        if not status:
            return [FileStatusOrNull(path=path) for path in _list]
        else:
            return [
                FileStatusOrNull(
                    path=st[0],
                    status=FileStatus(
                        fileId=str(st[1]["fileId"]),
                        length=str(st[1]["length"]),
                        type=str(st[1]["type"]),
                        modificationTime=str(st[1]["modificationTime"]),
                    ),
                )
                for st in _list
            ]

    def status(self, path, strict):
        try:
            status = self.fs.status(path, False)
        except Exception as e:
            raise FileSystemException(str(e))
        if status is None:
            if strict:
                raise FileSystemException("File %s does not exist" % path)
            else:
                return FileStatusOrNull(path=path)
        else:
            return FileStatusOrNull(
                path=path,
                status=FileStatus(
                    fileId=str(status["fileId"]),
                    length=str(status["length"]),
                    type=str(status["type"]),
                    modificationTime=str(status["modificationTime"]),
                ),
            )

    def content(self, path, strict):
        try:
            content = self.fs.content(path, False)
        except Exception as e:
            raise FileSystemException(str(e))
        if content is None:
            if strict:
                raise FileSystemException("File %s does not exist" % path)
            else:
                return FileContentOrNull(path=path)
        else:
            return FileContentOrNull(
                path=path,
                content=FileContent(
                    length=str(content["length"]),
                    fileCount=str(content["fileCount"]),
                    directoryCount=str(content["directoryCount"]),
                ),
            )

    def rm(self, path, recursive):
        try:
            self.fs.delete(path, recursive)
        except FileSystemError:
            raise FileSystemException("File %s does not exist" % path)
        except Exception as e:
            raise FileSystemException(str(e))

    def copy(self, src_path, dst_path):
        try:
            self.fs.copy(src_path, dst_path)
        except Exception as e:
            raise FileSystemException(str(e))

    def rename(self, src_path, dst_path):
        try:
            self.fs.rename(src_path, dst_path)
        except Exception as e:
            raise FileSystemException(str(e))

    def set_owner(self, path, owner, group):
        try:
            self.fs.set_owner(path, owner, group)
        except Exception as e:
            raise FileSystemException(str(e))

    def set_permission(self, path, permission):
        try:
            self.fs.set_permission(path, permission)
        except Exception as e:
            raise FileSystemException(str(e))

    def mkdir(self, path, permission):
        try:
            self.fs.mkdir(path, permission)
        except Exception as e:
            raise FileSystemException(str(e))

    def open(self, path, mode):
        if "b" not in mode:
            mode += "b"
        try:
            file = self.fs.open(path=path, mode=mode)
        except Exception as e:
            raise FileSystemException(str(e))
        desc = str(file.fileno())
        self.file_descs[desc] = file
        return desc

    def flush(self, filedesc):
        if filedesc not in self.file_descs:
            raise FileSystemException(
                "No such file descriptor %s, is the file open !" % filedesc
            )
        file = self.file_descs[filedesc]
        try:
            file.flush()
        except Exception as e:
            raise FileSystemException(str(e))

    def close(self, filedesc):
        if filedesc not in self.file_descs:
            raise FileSystemException(
                "No such file descriptor %s, is the file open !" % filedesc
            )
        file = self.file_descs[filedesc]
        try:
            file.close()
        except Exception as e:
            raise FileSystemException(str(e))
        self.file_descs.pop(filedesc)

    def read(self, filedesc, size):
        if filedesc not in self.file_descs:
            raise FileSystemException(
                "No such file descriptor %s, is the file open !" % filedesc
            )
        file = self.file_descs[filedesc]
        try:
            return file.read(size)
        except Exception as e:
            raise FileSystemException(str(e))

    def readline(self, filedesc):
        if filedesc not in self.file_descs:
            raise FileSystemException(
                "No such file descriptor %s, is the file open !" % filedesc
            )
        file = self.file_descs[filedesc]
        try:
            return file.readline()
        except Exception as e:
            raise FileSystemException(str(e))

    def readlines(self, filedesc):
        if filedesc not in self.file_descs:
            raise FileSystemException(
                "No such file descriptor %s, is the file open !" % filedesc
            )
        file = self.file_descs[filedesc]
        try:
            return file.readlines()
        except Exception as e:
            raise FileSystemException(str(e))

    def write(self, filedesc, data):
        if filedesc not in self.file_descs:
            raise FileSystemException(
                "No such file descriptor %s, is the file open !" % filedesc
            )
        file = self.file_descs[filedesc]
        try:
            pos = file.write(data)
            return pos if pos is not None else file.tell()
        except Exception as e:
            raise FileSystemException(str(e))

    def writelines(self, filedesc, lines):
        if filedesc not in self.file_descs:
            raise FileSystemException(
                "No such file descriptor %s, is the file open !" % filedesc
            )
        file = self.file_descs[filedesc]
        try:
            return file.writelines(lines)
        except Exception as e:
            raise FileSystemException(str(e))

    def tell(self, filedesc):
        if filedesc not in self.file_descs:
            raise FileSystemException(
                "No such file descriptor %s, is the file open !" % filedesc
            )
        file = self.file_descs[filedesc]
        try:
            return file.tell()
        except Exception as e:
            raise FileSystemException(str(e))

    def seek(self, filedesc, position):
        if filedesc not in self.file_descs:
            raise FileSystemException(
                "No such file descriptor %s, is the file open !" % filedesc
            )
        file = self.file_descs[filedesc]
        try:
            pos = file.seek(position)
            return pos if pos is not None else file.tell()
        except Exception as e:
            raise FileSystemException(str(e))

    def readable(self, filedesc):
        if filedesc not in self.file_descs:
            raise FileSystemException(
                "No such file descriptor %s, is the file open !" % filedesc
            )
        file = self.file_descs[filedesc]

        try:
            return file.readable()
        except AttributeError:
            # python 27
            try:
                file.seek(file.tell())
                return True
            except Exception:
                # non seekable
                return False
        except Exception as e:
            raise FileSystemException(str(e))

    def writable(self, filedesc):
        if filedesc not in self.file_descs:
            raise FileSystemException(
                "No such file descriptor %s, is the file open !" % filedesc
            )
        file = self.file_descs[filedesc]
        try:
            return file.writable()
        except AttributeError:
            # python 27
            if "w" in file.mode or "a" in file.mode or "+" in file.mode:
                return True
            else:
                return False
        except Exception as e:
            raise FileSystemException(str(e))

    def seekable(self, filedesc):
        if filedesc not in self.file_descs:
            raise FileSystemException(
                "No such file descriptor %s, is the file open !" % filedesc
            )
        file = self.file_descs[filedesc]
        try:
            return file.seekable()
        except AttributeError:
            # python 27
            if "w" in file.mode or "a" in file.mode or "+" in file.mode:
                return True
            else:
                return False
        except Exception as e:
            raise FileSystemException(str(e))



import logging as lg
from copy import deepcopy
from threading import RLock

from .config import FileSystemsConfig
from .gcp.filesystem import GCPFileSystem
from .hdfs.filesystem import HDFSFileSystem
from .ioutils import FileSystemError
from .local.local_filesystem import LocalFileSystem
from .local.remote_filesystem import RemoteFileSystem
from .s3.filesystem import S3FileSystem

_logger = lg.getLogger(__name__)

_glock = RLock()


class FileSystemFactory(object):
    __instance = None

    @staticmethod
    def getInstance():
        with _glock:
            """ Static access method. """
            if FileSystemFactory.__instance is None:
                FileSystemFactory()
        return FileSystemFactory.__instance

    def __init__(self):
        if FileSystemFactory.__instance is not None:
            raise Exception("Only one instance of Client Factory is allowed!")
        else:
            FileSystemFactory.__instance = self
            self._configure()

    def _configure(self):
        self._filesystems = {}
        _config = FileSystemsConfig(None).get_config()
        if _config is not None:
            for _fs_conf in _config["filesystems"]:
                self.register_filesystem(_fs_conf)
        else:
            _logger.warning("Empty filesystems configuration file.")

    def register_filesystem(self, conf):
        _conf = deepcopy(conf)
        with _glock:
            if "name" in _conf:
                name = _conf.pop("name")
                if name in self._filesystems:
                    # file system already exist, overwrite
                    del self._filesystems[name]
                self._filesystems[name] = _conf
                _logger.info("Registered filesystem '%s'" % name)
            else:
                raise InvalidFileSystemError(
                    "Missing filesystem name from %s" % str(_conf)
                )

    def get_filesystem(self, name):
        def _get_filesystem(name):
            if name in self._filesystems:
                return deepcopy(self._filesystems[name])
            else:
                # the name does not exist
                raise UnknownFileSystemError("Unknown Filesystem '%s'" % name)

        with _glock:
            config = _get_filesystem(name)
            if "type" not in config:
                raise InvalidFileSystemError("filesystem type missing for '%s'" % name)
            else:
                fs_type = config.pop("type")
                if fs_type == "local":
                    if "address" in config and "port" in config:
                        address = config["address"]
                        port = int(config["port"])
                        filesystem = RemoteFileSystem(address, port)
                    else:
                        filesystem = LocalFileSystem()
                elif fs_type == "hdfs":
                    auth_mechanism = config["auth_mechanism"]
                    del config["auth_mechanism"]
                    filesystem = HDFSFileSystem(name, auth_mechanism, **config)
                elif fs_type == "s3":
                    bucket = config["bucket"]
                    del config["bucket"]
                    filesystem = S3FileSystem(bucket, **config)
                elif fs_type == "gcs":
                    bucket = config["bucket"]
                    del config["bucket"]
                    if "token" in config:
                        token = config["token"]
                        del config["token"]
                    else:
                        token = None
                    filesystem = GCPFileSystem(bucket, token, **config)
                else:
                    raise NonSupportedFileSystemError(
                        "Non supported filesystem type '%s'" % type
                    )
            # This is an ugly workaround to initialize the client
            filesystem.list("/")
        return filesystem


class NonSupportedFileSystemError(FileSystemError):
    pass


class InvalidFileSystemError(FileSystemError):
    pass


class UnknownFileSystemError(FileSystemError):
    pass



import io
import json
import logging as lg
import os
import re
import time

import google.auth
from google.cloud import storage
from google.oauth2 import service_account
from google.oauth2.credentials import Credentials

from ...common.utils.glob import iglob
from ..filesystem import IFileSystem
from ..ioutils import FileSystemError

_logger = lg.getLogger(__name__)


class GCPFileSystem(IFileSystem):
    def __init__(self, bucket_name, token=None, **params):
        credentials = self.connect(token=token)
        self.client = storage.Client(credentials=credentials, **params)
        self.bucket = self.client.get_bucket(bucket_name)

    def connect(self, token):
        """Connect  using a concrete token.

        Parameters
        ----------
        token: str, dict or Credentials
            If a str, try to load as a Service file, or next as a JSON; if
            dict, try to interpret as credentials; if Credentials, use directly.
        """

        def _connect_service(service_file):
            # raises exception if file does not match expectation
            return service_account.Credentials.from_service_account_file(service_file)

        def _dict_to_credentials(token):
            try:
                credentials = service_account.Credentials.from_service_account_info(
                    token
                )
            except Exception:
                credentials = Credentials(
                    None,
                    refresh_token=token["refresh_token"],
                    client_secret=token["client_secret"],
                    client_id=token["client_id"],
                    token_uri="https://oauth2.googleapis.com/token",
                )
            return credentials

        if token is None:
            return None
        if isinstance(token, str):
            if not os.path.exists(token):
                raise FileNotFoundError(token)
            try:
                # is this a "service" token?
                return _connect_service(token)
            except Exception:
                token = json.load(open(token))
        if isinstance(token, dict):
            return _dict_to_credentials(token)
        if isinstance(token, google.auth.credentials.Credentials):
            return token
        else:
            raise ValueError("Token format not understood")

    def resolve_path(self, path):
        # Remove any leading and trailing slash
        rpath = os.path.normpath(re.sub("^/*", "", path))
        return "" if rpath == "." else rpath

    def format_path(self, path):
        if path.startswith(self.bucket.name + "/"):
            return re.sub("^%s" % self.bucket.name, "", path)
        else:
            if not path.startswith("/"):
                return "/" + path

    def _status(self, path, strict=True):
        def _datetime2timestamp(dt):
            try:
                int(dt.timestamp() * 1000)
            except AttributeError:
                # python 2.7
                int(time.mktime(dt.timetuple()) * 1000)

        if path in ["", "/"]:
            # consider empty or slash the root of the bucket
            return {
                "fileId": "/",
                "type": "DIRECTORY",
                "length": 0,
                "creationTime": None,
                "modificationTime": None,
            }
        # check if the path is a prefix
        if len(self._list_prefix(path)) != 0:
            return {
                "fileId": self.bucket.name + "/" + path,
                "type": "DIRECTORY",
                "length": 0,
                "creationTime": None,
                "modificationTime": None,
            }
        # check if the path is a real object
        blob = self.bucket.get_blob(path)
        if blob is not None:
            return {
                "fileId": blob.id,
                "type": "FILE",
                "length": blob.size,
                "creationTime": _datetime2timestamp(blob.time_created),
                "modificationTime": _datetime2timestamp(blob.updated),
            }
        # the path does not exist
        if not strict:
            return None
        else:
            raise FileSystemError("%r does not exist." % path)

    def _list_prefix(self, prefix):
        if prefix not in [".", ""]:
            # make sure the prefix have a tailing slash
            pfx = prefix + "/" if not str(prefix).endswith("/") else prefix
        else:
            pfx = ""
        iter = self.bucket.list_blobs(delimiter="/", prefix=pfx)
        files = [os.path.basename(blob.name) for blob in iter]
        # prefixes have a tailing slash, make sure to strip it
        files.extend(
            [os.path.basename(os.path.normpath(prefix)) for prefix in iter.prefixes]
        )
        return [file for file in files if file not in ["", ".", "/"]]

    def _list(self, path, status=False, glob=False):
        if not glob:
            files = self._list_prefix(path)
            if len(files) == 0:
                raise FileSystemError("%r is not a directory." % path)
            elif len(files) == 1 and files[0] == "":
                # special case when the directory exist and is empty
                # _list returns [''] return empty array instead
                return []
            else:
                if status:
                    return [(f, self.status(os.path.join(path, f))) for f in files]
                else:
                    return files
        else:
            files = [i_file for i_file in iglob(self, path)]
            if status:
                return [(f, self.status(f)) for f in files]
            else:
                return files

    def _delete(self, path, recursive=False):
        for blob in self.bucket.list_blobs(prefix=path):
            blob.delete()

    def _copy(self, src_path, dst_path):
        src_blob = self.bucket.get_blob(src_path)
        self.bucket.copy_blob(
            src_blob, self.bucket, re.sub("^" + src_path, dst_path, src_blob.name)
        )

    def _set_owner(self, path, owner=None, group=None):
        raise NotImplementedError

    def _set_permission(self, path, permission):
        raise NotImplementedError

    def _mkdir(self, path, permission=None):
        blob = self.bucket.blob(path + "/")
        blob.upload_from_string("")

    def _open(
        self,
        path,
        mode="rb",
        buffer_size=1024,
        part_size=5 * 2 ** 20,
        acl="",
        encoding=None,
        **kwargs
    ):
        mode2 = mode if "b" in mode else (mode.replace("t", "") + "b")
        raw_file = GCSFile(
            self.bucket,
            path,
            mode2,
            buffer_size=buffer_size,
            part_size=part_size,
            acl=acl,
        )
        if "b" in mode:
            return raw_file
        return io.TextIOWrapper(raw_file, encoding=encoding)


class GCSFile(object):
    def __init__(
        self, bucket, path, mode="rb", buffer_size=1024, part_size=16 * 2 ** 20, acl=""
    ):
        self.mode = mode
        if mode not in {"rb", "wb", "ab"}:
            raise NotImplementedError(
                "File mode must be {'rb', 'wb', 'ab'}, " "not %s" % mode
            )
        self.path = path
        self.bucket = bucket
        self.part_size = part_size
        self.loc = 0  # The read/write location pointer
        self.closed = False
        self.acl = acl
        if self.writable():
            self.buffer = io.BytesIO()
            self.size = 0
            self.forced = False
            if "a" in mode:
                self.blob = self.bucket.get_blob(self.path)
                if self.blob is None:
                    raise FileSystemError(
                        "Can not append to non existent file %s" % self.path
                    )
            else:
                self.blob = self.bucket.blob(self.path)
                # create the actual blob
                self.blob.upload_from_string("")
        else:
            self.cache = b""
            self.buffer_size = buffer_size
            self.start = None  # The start position of the cache
            self.end = None  # The end position of the cache
            self.blob = self.bucket.get_blob(self.path)
            if self.blob is None:
                raise FileSystemError("Can not read non existent file %s" % self.path)
            self.size = self.blob.size

    def tell(self):
        """Tell Current file location."""
        return self.loc

    def seek(self, loc, whence=0):
        """Set current file location.

        Parameters
        ----------
        loc : int
            byte location
        whence : {0, 1, 2}
            from start of file, current location or end of file, resp.
        """
        if not self.readable():
            raise ValueError("Seek only available in read mode")
        if whence == 0:
            nloc = loc
        elif whence == 1:
            nloc = self.loc + loc
        elif whence == 2:
            nloc = self.size + loc
        else:
            raise ValueError("invalid whence (%s, should be 0, 1 or 2)" % whence)
        if nloc < 0:
            raise ValueError("Seek before start of file")
        self.loc = nloc
        return self.loc

    def readline(self, length=-1):
        """Read and return a line from the stream.

        If length is specified, at most size bytes will be read.
        """
        self._fetch(self.loc, self.loc + 1)
        while True:
            found = self.cache[self.loc - self.start :].find(b"\n") + 1  # NOOP
            if 0 < length < found:
                return self.read(length)
            if found:
                return self.read(found)
            if self.end > self.size:
                return self.read(length)
            self._fetch(self.start, self.end + self.buffer_size)

    def __next__(self):
        out = self.readline()
        if not out:
            raise StopIteration
        return out

    next = __next__

    def __iter__(self):
        return self

    def readlines(self):
        """Return all lines in a file as a list."""
        return list(self)

    def _fetch(self, start, end):
        part_end = max(end, start + self.buffer_size)
        if self.start is None and self.end is None:
            # First read
            self.start = start
            self.end = part_end
            _logger.debug("fetch %s --> %s" % (self.start, self.end))
            self.cache = self.blob.download_as_bytes(start=self.start, end=self.end)
        if start < self.start:
            if part_end < self.start:
                self.start, self.end = None, None
                return self._fetch(start, end)
            else:
                _logger.debug("fetch %s --> %s" % (start, self.start))
                pre = self.blob.download_as_bytes(start=start, end=self.start)
                self.start = start
                if part_end < self.end:
                    self.cache = pre + self.cache[self.start : part_end]
                    self.end = part_end
                else:
                    self.cache = pre + self.cache
        if end > self.end:
            if self.end > self.size:
                return
            if start > self.end:
                self.start, self.end = None, None
                return self._fetch(start, end)
            else:
                _logger.debug("fetch %s --> %s" % (self.end, part_end))
                new = self.blob.download_as_bytes(start=self.end, end=part_end)
                self.end = part_end
                if start <= self.start:
                    self.cache = self.cache + new
                else:
                    self.cache = self.cache[start:] + new
                    self.start = start

    def read(self, length=-1):
        """
        Return data from cache, or fetch pieces as necessary.

        Parameters
        ----------
        length : int (-1)
            Number of bytes to read; if <0, all remaining bytes.
        """
        if not self.readable():
            raise ValueError("File not in read mode")
        if length < 0:
            length = self.size
        if self.closed:
            raise ValueError("I/O operation on closed file.")
        self._fetch(self.loc, self.loc + length)
        out = self.cache[self.loc - self.start : self.loc - self.start + length]
        self.loc += len(out)
        if len(out) > 0:
            return out
        else:
            return None

    def write(self, data):
        """Write data to a the GCS file.

        Writes the data to a temporary file and flush to remote storage on close()
        or when file size reach part_size.
        Parameters
        ----------
        data : bytes
            Set of bytes to be written.
        """
        if not self.writable():
            raise ValueError("File not in write mode")
        if self.closed:
            raise ValueError("I/O operation on closed file.")
        out = self.buffer.write(data)
        self.loc += out
        if self.buffer.tell() >= self.part_size:
            # flush the buffer into a part
            self.flush()
        return out

    def flush(self, force=False):
        """Write buffered data to GCS.

        Uploads the current FILE, if it is larger than the part-size. If
        the buffer is smaller than the  part-size, this is a no-op.
        Parameters
        ----------
        force : bool
            When closing, write the last block even if it is smaller than
            blocks are allowed to be.
        """
        if self.writable() and not self.closed:
            if self.buffer.tell() < self.part_size and not force:
                # ignore if not enough data for a block and not closing
                return
            if self.buffer.tell() == 0:
                # no data in the buffer to write
                return
            if force and self.forced:
                raise ValueError("Force flush cannot be called more than once")
            if force:
                self.forced = True

            self.buffer.seek(0)
            part_blob = self.bucket.blob(self.path + ".part")
            part_blob.upload_from_file(self.buffer)

            self.blob.compose([self.blob, part_blob])
            part_blob.delete()

            self.buffer = io.BytesIO()

    def close(self):
        """Close file.

        If in write mode, key is only finalized upon close, and key will then
        be available to other processes.
        """
        if self.closed:
            return
        self.cache = None
        if self.writable():
            self.flush(force=True)
            self.buffer.seek(0)
            self.buffer.truncate(0)
        self.closed = True

    def readable(self):
        """Return whether the GCSFile was opened for reading."""
        return "r" in self.mode

    def seekable(self):
        """Return whether the GCSFile can be seeked (only in read mode)."""
        return self.readable()

    def writable(self):
        """Return whether the GCSFile was opened for writing."""
        return "w" in self.mode or "a" in self.mode

    def __del__(self):
        self.close()

    def __str__(self):
        return "<GCSFile %s/%s>" % (self.bucket, self.path)

    __repr__ = __str__

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()



import io
import logging as lg
import time

from thrift.protocol import TBinaryProtocol
from thrift.transport import TSocket
from thrift.transport import TTransport

from ...thrift.worker.filesystem import LocalFilesystem

_logger = lg.getLogger(__name__)


class LocalFileSystemClient(object):
    def __init__(self, host="127.0.0.1", port=8989):
        self.host = host
        self.port = port
        # Init thrift connection and protocol handlers
        socket = TSocket.TSocket(host, port)
        self.transport = TTransport.TBufferedTransport(socket)
        protocol = TBinaryProtocol.TBinaryProtocol(self.transport)

        # Set client to our Example
        self.client = LocalFilesystem.Client(protocol)

    def start(self):
        # Connect to server
        retries = 1
        last_error = None
        while retries < 10:
            try:
                self.transport.open()
                break
            except TTransport.TTransportException as e:
                _logger.error(
                    "Could not connect to any local system service. "
                    + "retrying after 5 seconds ..."
                )
                last_error = e
            retries += 1
            time.sleep(5.0)

        if retries == 10:
            _logger.error(
                "Could not connect to any local system service after 30 retries. "
                + "exiting ..."
            )
            self.stop()
            raise last_error

    def ls(self, path, status=False, glob=False):
        if status:
            return [
                (
                    st.path,
                    {
                        "fileId": st.status.fileId,
                        "length": st.status.length,
                        "type": st.status.type,
                        "modificationTime": st.status.modificationTime,
                    },
                )
                for st in self.client.ls(path, status, glob)
            ]
        else:
            return [status.path for status in self.client.ls(path, status, glob)]

    def status(self, path, strict=True):
        status = self.client.status(path, strict)
        if status.status:
            return {
                "fileId": status.status.fileId,
                "length": status.status.length,
                "type": status.status.type,
                "modificationTime": status.status.modificationTime,
            }
        else:
            return None

    def content(self, path, strict=True):
        content = self.client.content(path, strict)
        if content.content:
            return {
                "length": content.content.length,
                "fileCount": content.content.fileCount,
                "directoryCount": content.content.directoryCount,
            }
        else:
            return None

    def rm(self, path, recursive=False):
        self.client.rm(path, recursive)

    def rename(self, src_path, dst_path):
        self.client.rename(src_path, dst_path)

    def copy(self, src_path, dst_path):
        self.client.copy(src_path, dst_path)

    def set_owner(self, path, owner=None, group=None):
        self.client.set_owner(path, owner, group)

    def set_permission(self, path, permission):
        self.client.set_permission(path, permission)

    def mkdir(self, path, permission):
        self.client.mkdir(path, permission)

    def open(self, path, mode, buffer_size=1024, encoding=None):
        mode2 = mode if "b" in mode else (mode.replace("t", "") + "b")
        file = LocalFile(self.client, path, mode2)
        if file.readable() and file.writable():
            file = io.BufferedRandom(file, buffer_size)
        elif file.readable():
            file = io.BufferedReader(file, buffer_size)
        elif file.writable():
            file = io.BufferedWriter(file, buffer_size)
        if encoding:
            return io.TextIOWrapper(file, encoding=encoding)
        else:
            return file

    def stop(self):
        self.transport.close()


class LocalFile(object):
    def __init__(self, client, path, mode="rb"):
        if "b" not in mode:
            raise NotImplementedError("Only binary read/write modes are supported.")
        self.client = client
        self.loc = 0  # The read/write location pointer
        self.closed = False
        self.desc = self.client.open(path=path, mode=mode)

    def tell(self):
        return self.client.tell(self.desc)

    def seek(self, loc, whence=0):
        return self.client.seek(self.desc, loc)

    def read(self, size=-1):
        return self.client.read(self.desc, size)

    def readline(self):
        return self.client.readline(self.desc)

    def readlines(self, lines):
        return self.client.readlines(self.desc, lines)

    def write(self, data):
        return self.client.write(self.desc, data)

    def writelines(self, lines):
        return self.client.writelines(self.desc, lines)

    def seekable(self):
        return self.client.seekable(self.desc)

    def readable(self):
        return self.client.readable(self.desc)

    def writable(self):
        return self.client.writable(self.desc)

    def flush(self):
        self.client.flush(self.desc)

    def close(self):
        if self.closed:
            return
        self.closed = True
        self.client.close(self.desc)

    def __str__(self):
        return "<RemoteFile %s:%s %s>" % (self.client.host, self.client.port, self.path)

    __repr__ = __str__

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()

    # Implementations of BufferedIOBase stub methods

    def read1(self, length=-1):
        return self.read(length)

    def detach(self):
        raise io.UnsupportedOperation()

    def readinto(self, b):
        data = self.client.read(self.desc, len(b))
        b[: len(data)] = bytes(data)
        return len(data)

    def readinto1(self, b):
        return self.readinto(b)



import logging as lg
import time
from threading import Thread

_logger = lg.getLogger(__name__)


class SimpleScheduler(object):
    """Manages Job objects and the waiting between job executions."""

    def __init__(self, lqueue, cqueue, callback=None):
        self.cqueue = cqueue
        self.lqueue = lqueue
        self.callback = callback
        self.stopped = False

    def schedule(self, task):
        self.lqueue.put(task)

    def _run(self):
        while True:
            if not self.lqueue.empty():
                task_exec = self.lqueue.get()
                _logger.debug("Scheduling next task %s for execution.", task_exec.tid)
                if self.callback is not None:
                    self.callback(task_exec.tid)
                self.cqueue.put(task_exec)
                self.lqueue.task_done()
            else:
                _logger.debug("No new tasks to schedule, sleeping for %s seconds...", 2)
                time.sleep(2)
            if self.stopped and self.lqueue.empty():
                _logger.debug("No new tasks to schedule, terminating scheduler thread.")
                return

    def start(self):
        _logger.info("Stating tanit scheduler.")
        self.daemon = Thread(target=self._run, args=())
        self.daemon.setDaemon(True)
        self.daemon.start()
        _logger.info("Tanit scheduler started.")

    def stop(self):
        _logger.info("Stopping tanit scheduler.")
        self.stopped = True
        self.daemon.join()
        _logger.info("Tanit scheduler stopped.")



import logging as lg

from ...common.config.config import Config
from ...common.config.config import TanitConfigurationException

_logger = lg.getLogger(__name__)


class MasterConfig(Config):
    default_bind_address = "0.0.0.0"
    default_thrift_threads = 25

    def __init__(self, path=None):
        super(MasterConfig, self).__init__(path)

    def load(self):

        if self.config.has_option("master", "bind_address"):
            self.bind_address = self.config.get("master", "bind_address")
        else:
            self.bind_address = MasterConfig.default_bind_address

        if self.config.has_option("master", "worker-service_address"):
            worker_service_address = self.config.get("master", "worker-service_address")
        else:
            raise TanitConfigurationException(
                "Missing master rpc worker service address from configuration"
            )

        if len(worker_service_address.split(":")) != 2:
            raise TanitConfigurationException(
                "Master rpc worker service address should be formatted as host:port."  # NOQA
            )

        self.worker_service_host = worker_service_address.split(":")[0]
        self.worker_service_port = int(worker_service_address.split(":")[1])

        if self.config.has_option("master", "client-service_address"):
            client_service_address = self.config.get("master", "client-service_address")
        else:
            raise TanitConfigurationException(
                "Missing master rpc client service address from configuration"
            )

        if len(client_service_address.split(":")) != 2:
            raise TanitConfigurationException(
                "Master rpc client service address should be formatted as host:port."  # NOQA
            )

        self.client_service_host = client_service_address.split(":")[0]
        self.client_service_port = int(client_service_address.split(":")[1])

        if self.config.has_option("worker", "thrift_threads"):
            self.thrift_threads = self.config.getint("worker", "thrift_threads")
        else:
            self.thrift_threads = MasterConfig.default_thrift_threads



#
# Autogenerated by Thrift Compiler (0.13.0)
#
# DO NOT EDIT UNLESS YOU ARE SURE THAT YOU KNOW WHAT YOU ARE DOING
#
#  options string: py
#

from thrift.Thrift import TType, TMessageType, TFrozenDict, TException, TApplicationException
from thrift.protocol.TProtocol import TProtocolException
from thrift.TRecursive import fix_spec

import sys

from thrift.transport import TTransport
all_structs = []


class FileStatus(object):
    """
    Attributes:
     - fileId
     - length
     - type
     - modificationTime

    """


    def __init__(self, fileId=None, length=None, type=None, modificationTime=None,):
        self.fileId = fileId
        self.length = length
        self.type = type
        self.modificationTime = modificationTime

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRING:
                    self.fileId = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.STRING:
                    self.length = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 3:
                if ftype == TType.STRING:
                    self.type = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 4:
                if ftype == TType.STRING:
                    self.modificationTime = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('FileStatus')
        if self.fileId is not None:
            oprot.writeFieldBegin('fileId', TType.STRING, 1)
            oprot.writeString(self.fileId.encode('utf-8') if sys.version_info[0] == 2 else self.fileId)
            oprot.writeFieldEnd()
        if self.length is not None:
            oprot.writeFieldBegin('length', TType.STRING, 2)
            oprot.writeString(self.length.encode('utf-8') if sys.version_info[0] == 2 else self.length)
            oprot.writeFieldEnd()
        if self.type is not None:
            oprot.writeFieldBegin('type', TType.STRING, 3)
            oprot.writeString(self.type.encode('utf-8') if sys.version_info[0] == 2 else self.type)
            oprot.writeFieldEnd()
        if self.modificationTime is not None:
            oprot.writeFieldBegin('modificationTime', TType.STRING, 4)
            oprot.writeString(self.modificationTime.encode('utf-8') if sys.version_info[0] == 2 else self.modificationTime)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class FileStatusOrNull(object):
    """
    Attributes:
     - path
     - status

    """


    def __init__(self, path=None, status=None,):
        self.path = path
        self.status = status

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRING:
                    self.path = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.STRUCT:
                    self.status = FileStatus()
                    self.status.read(iprot)
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('FileStatusOrNull')
        if self.path is not None:
            oprot.writeFieldBegin('path', TType.STRING, 1)
            oprot.writeString(self.path.encode('utf-8') if sys.version_info[0] == 2 else self.path)
            oprot.writeFieldEnd()
        if self.status is not None:
            oprot.writeFieldBegin('status', TType.STRUCT, 2)
            self.status.write(oprot)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class FileContent(object):
    """
    Attributes:
     - length
     - fileCount
     - directoryCount

    """


    def __init__(self, length=None, fileCount=None, directoryCount=None,):
        self.length = length
        self.fileCount = fileCount
        self.directoryCount = directoryCount

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRING:
                    self.length = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.STRING:
                    self.fileCount = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 3:
                if ftype == TType.STRING:
                    self.directoryCount = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('FileContent')
        if self.length is not None:
            oprot.writeFieldBegin('length', TType.STRING, 1)
            oprot.writeString(self.length.encode('utf-8') if sys.version_info[0] == 2 else self.length)
            oprot.writeFieldEnd()
        if self.fileCount is not None:
            oprot.writeFieldBegin('fileCount', TType.STRING, 2)
            oprot.writeString(self.fileCount.encode('utf-8') if sys.version_info[0] == 2 else self.fileCount)
            oprot.writeFieldEnd()
        if self.directoryCount is not None:
            oprot.writeFieldBegin('directoryCount', TType.STRING, 3)
            oprot.writeString(self.directoryCount.encode('utf-8') if sys.version_info[0] == 2 else self.directoryCount)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class FileContentOrNull(object):
    """
    Attributes:
     - path
     - content

    """


    def __init__(self, path=None, content=None,):
        self.path = path
        self.content = content

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRING:
                    self.path = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.STRUCT:
                    self.content = FileContent()
                    self.content.read(iprot)
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('FileContentOrNull')
        if self.path is not None:
            oprot.writeFieldBegin('path', TType.STRING, 1)
            oprot.writeString(self.path.encode('utf-8') if sys.version_info[0] == 2 else self.path)
            oprot.writeFieldEnd()
        if self.content is not None:
            oprot.writeFieldBegin('content', TType.STRUCT, 2)
            self.content.write(oprot)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class FileSystemException(TException):
    """
    Attributes:
     - message

    """


    def __init__(self, message=None,):
        self.message = message

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRING:
                    self.message = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('FileSystemException')
        if self.message is not None:
            oprot.writeFieldBegin('message', TType.STRING, 1)
            oprot.writeString(self.message.encode('utf-8') if sys.version_info[0] == 2 else self.message)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __str__(self):
        return repr(self)

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(FileStatus)
FileStatus.thrift_spec = (
    None,  # 0
    (1, TType.STRING, 'fileId', 'UTF8', None, ),  # 1
    (2, TType.STRING, 'length', 'UTF8', None, ),  # 2
    (3, TType.STRING, 'type', 'UTF8', None, ),  # 3
    (4, TType.STRING, 'modificationTime', 'UTF8', None, ),  # 4
)
all_structs.append(FileStatusOrNull)
FileStatusOrNull.thrift_spec = (
    None,  # 0
    (1, TType.STRING, 'path', 'UTF8', None, ),  # 1
    (2, TType.STRUCT, 'status', [FileStatus, None], None, ),  # 2
)
all_structs.append(FileContent)
FileContent.thrift_spec = (
    None,  # 0
    (1, TType.STRING, 'length', 'UTF8', None, ),  # 1
    (2, TType.STRING, 'fileCount', 'UTF8', None, ),  # 2
    (3, TType.STRING, 'directoryCount', 'UTF8', None, ),  # 3
)
all_structs.append(FileContentOrNull)
FileContentOrNull.thrift_spec = (
    None,  # 0
    (1, TType.STRING, 'path', 'UTF8', None, ),  # 1
    (2, TType.STRUCT, 'content', [FileContent, None], None, ),  # 2
)
all_structs.append(FileSystemException)
FileSystemException.thrift_spec = (
    None,  # 0
    (1, TType.STRING, 'message', 'UTF8', None, ),  # 1
)
fix_spec(all_structs)
del all_structs



import logging as lg
import time
from threading import Thread

from ..execution.execution_state import ExecutionState

_logger = lg.getLogger(__name__)


class WorkerDecommissioner(Thread):
    def __init__(self, execution_manager, workers_manager):
        super(WorkerDecommissioner, self).__init__()
        self.execution_manager = execution_manager
        self.workers_manager = workers_manager
        self.stopped = False

    def run(self):
        while not self.stopped:
            for worker in self.workers_manager.list_decommissioning_workers():
                self._decommission_worker(worker)
                self.workers_manager.on_worker_decommissioned(worker.wid)

    def stop(self):
        self.stopped = True

    def _decommission_worker(self, worker):
        worker_tasks = self.execution_manager.get_tasks(worker=worker.wid)

        while not self.stopped:
            # While the worker is up, wait for the tasks to finish
            # check if the worker is still alive
            try:
                worker.status()
            except Exception:
                break
            # check how many tasks are still alive
            worker_tasks = [
                task
                for task in worker_tasks
                if task.state not in [ExecutionState.FINISHED, ExecutionState.FAILED]
            ]
            if len(worker_tasks) == 0:
                break
            else:
                time.sleep(1.0)

        # check if all tasks actually
        if len(worker_tasks) == 0:
            _logger.info("Worker [ %s ] have no more active tasks.", worker.wid)
        else:
            _logger.info(
                "Worker [ %s ] still have %s active tasks.",
                worker.wid,
                len(worker_tasks),
            )
            # reschedule the tasks
            for task_exec in worker_tasks:
                self.execution_manager.task_reset(task_exec.tid)



import logging as lg
from contextlib import contextmanager

from pywhdfs.client import create_client

from ..filesystem import IFileSystem

_logger = lg.getLogger(__name__)


class HDFSFileSystem(IFileSystem):
    def __init__(self, cluster, auth_mechanism, **params):
        self.cluster = cluster
        self.client = create_client(auth_mechanism, **params)

    def resolve_path(self, path):
        return self.client.resolve_path(path)

    def _list(self, path, status=False, glob=False):
        return self.client.list(path, status, glob)

    def _status(self, path, strict=True):
        return self.client.status(path, strict)

    def _content(self, path, strict=True):
        c = self.client.content(path, strict)
        if c is None:
            return None
        else:
            return {
                "length": c["length"],
                "fileCount": c["fileCount"],
                "directoryCount": c["directoryCount"],
            }

    def _delete(self, path, recursive=False):
        return self.client.delete(path, recursive)

    def _copy(self, src_path, dst_path):
        # HDFS does not support copy natively : ugly implementation
        with self.client.read(src_path, chunk_size=1024) as reader:
            self.client.write(dst_path, reader)

    def rename(self, src_path, dst_path):
        return self.client.rename(src_path, dst_path)

    def _set_owner(self, path, owner=None, group=None):
        return self.client.set_owner(path, owner, group)

    def _set_permission(self, path, permission):
        return self.client.set_permission(path, permission)

    def _mkdir(self, path, permission=None):
        return self.client.makedirs(path, permission)

    def _open(self, path, mode, buffer_size=-1, encoding=None, **kwargs):
        # HDFS library does not implement open method
        raise NotImplementedError

    @contextmanager
    def read(
        self,
        path,
        offset=0,
        buffer_size=1024,
        encoding=None,
        chunk_size=None,
        delimiter=None,
        **kwargs
    ):
        with self.client.read(
            path,
            offset=offset,
            buffer_size=buffer_size,
            encoding=encoding,
            chunk_size=chunk_size,
            delimiter=delimiter,
            **kwargs
        ) as reader:
            yield reader

    def write(
        self,
        path,
        data=None,
        overwrite=False,
        permission=None,
        buffer_size=1024,
        append=False,
        encoding=None,
        **kwargs
    ):
        return self.client.write(
            path,
            data=data,
            overwrite=overwrite,
            permission=permission,
            buffersize=buffer_size,
            append=append,
            encoding=encoding,
            **kwargs
        )



import logging as lg

from ...common.model.execution_type import ExecutionType
from ...common.model.task import Task
from ...thrift.worker.service.ttypes import TaskType
from ...thrift.worker.service.ttypes import WorkerStatus

_logger = lg.getLogger(__name__)


class WorkerServiceHandler(object):
    def __init__(self, worker):
        self.worker = worker

    def submit(self, task):

        if task.type == TaskType.COPY:
            etype = ExecutionType.COPY
        elif task.type == TaskType.UPLOAD:
            etype = ExecutionType.UPLOAD
        elif task.type == TaskType.MOCK:
            etype = ExecutionType.MOCK
        else:
            # should raise exception here
            pass

        self.worker.submit(Task(task.tid, etype, task.params))

    def worker_status(self):
        status = self.worker.get_stats()
        return WorkerStatus(
            status.wid, status.running, status.pending, status.available
        )

    def register_filesystem(self, filesystem):
        self.worker.register_filesystem(filesystem.name, filesystem.parameters)



import abc
import hashlib
import logging as lg
import os
import posixpath as psp
import types
from contextlib import contextmanager

import six

from .ioutils import ChunkFileReader
from .ioutils import DelimitedFileReader
from .ioutils import FileReader
from .ioutils import FileSystemError

_logger = lg.getLogger(__name__)


@six.add_metaclass(abc.ABCMeta)
class IFileSystem:
    def resolve_path(self, path):
        """Return absolute, normalized path.

        :param path: Remote path.
        """
        return path

    def format_path(self, path):
        """Return the path in the normalized display format.

        :param path: Remote path.
        """
        return path

    @abc.abstractmethod
    def _list(self, path, status=False, glob=False):
        """Return names of files contained in a remote folder.

        :param path: Remote path to a directory. If `path` doesn't exist
          or points to a normal file, an :class:`FileSystemError` will be raised.
        :glob: Whether the path should be considered a glob expressions
        :param status: Also return each file's corresponding FileStatus.
        """
        raise NotImplementedError

    def list(self, path, status=False, glob=False):
        """Return names of files contained in a remote folder.

        :param path: Remote path to a directory. If `path` doesn't exist
          or points to a normal file, an :class:`FileSystemError` will be raised.
        :glob: Whether the path should be considered a glob expressions
        :param status: Also return each file's corresponding FileStatus.
        """
        rpath = self.resolve_path(path)
        if not glob and not self.exists(rpath):
            raise FileSystemError("%r does not exist.", rpath)
        else:
            return self._list(rpath, status=status, glob=glob)

    @abc.abstractmethod
    def _status(self, path, strict=True):
        raise NotImplementedError

    def status(self, path, strict=True):
        """Get FileStatus_ for a file or folder on HDFS.

        :param path: path.
        :param strict: If `False`, return `None` rather than raise an exception if
          the path doesn't exist.
        """
        rpath = self.resolve_path(path)
        s = self._status(rpath, strict)
        if s is None:
            return None
        else:
            return {
                "fileId": rpath,
                "length": str(s["length"]),
                "type": str(s["type"]).upper(),
                "modificationTime": s["modificationTime"],
            }

    def exists(self, path):
        rpath = self.resolve_path(path)
        return self.status(rpath, strict=False) is not None

    def content(self, path, strict=True):
        def _get_size(start_path="."):
            total_folders = 0
            total_files = 0
            total_size = 0

            for dirpath, dirnames, filenames in self.walk(start_path):
                total_folders += len(dirnames)
                for f in filenames:
                    total_files += 1
                    total_size += int(self.status(os.path.join(dirpath, f))["length"])

            return {
                "length": str(total_size),
                "fileCount": str(total_files),
                "directoryCount": str(total_folders),
            }

        _logger.debug("Fetching content summary for %r.", path)
        rpath = self.resolve_path(path)
        if not self.exists(rpath):
            if not strict:
                return None
            else:
                raise FileSystemError("%r does not exist.", rpath)
        else:
            return _get_size(rpath)

    @abc.abstractmethod
    def _delete(self, path, recursive=True):
        raise NotImplementedError

    def delete(self, path, recursive=True):
        """Remove a file or directory.

        :param path: path.
        :param recursive: Recursively delete files and directories. By default,
          this method will raise an :class:`FileSystemError` if trying to delete a
          non-empty directory.
        This function returns `True` if the deletion was successful and `False` if
        no file or directory previously existed at `path`.
        """
        rpath = self.resolve_path(path)
        if not self.exists(rpath):
            raise FileSystemError("%r does not exist.", rpath)
        self._delete(rpath, recursive=recursive)

    @abc.abstractmethod
    def _copy(self, src_path, dst_path):
        raise NotImplementedError

    def copy(self, src_path, dst_path, recursive=True):
        """Copy a file.

        :param src_path: Source path.
        :param dst_path: Destination path. If the path already exists and is
          a directory, the source will be moved into it. If the path exists and is
          a file, or if a parent destination directory is missing, this method will
          raise an :class:`FileSystemError`.
        :param recursive: recursively copy.
        """
        rsrc_path = self.resolve_path(src_path)
        rdst_path = self.resolve_path(dst_path)
        src_status = self.status(rsrc_path, strict=False)
        dst_status = self.status(rdst_path, strict=False)
        if src_status is None:
            raise FileSystemError("%r does not exist.", rsrc_path)

        if src_status["type"] == "DIRECTORY":
            if not recursive:
                raise FileSystemError("'%s' is a directory." % src_path)
            if dst_status is None:
                self.mkdir(rdst_path)
            for rpath, _, files in self.walk(src_path):
                for f in files:
                    self._copy(os.path.join(rsrc_path, f), os.path.join(rdst_path, f))
        else:
            self._copy(rsrc_path, rdst_path)

    def rename(self, src_path, dst_path):
        """Rename a file or folder.

        Not all filesystems have a native rename implementation based
        on metadata only modification.
        :param src_path: Source path.
        :param dst_path: Destination path. If the path already exists and is
          a directory, the source will be moved into it. If the path exists and is
          a file, or if a parent destination directory is missing, this method will
          raise an :class:`FileSystemError`.
        """
        rsrc_path = self.resolve_path(src_path)
        rdst_path = self.resolve_path(dst_path)
        self.copy(rsrc_path, rdst_path, recursive=True)
        self.delete(rsrc_path)

    @abc.abstractmethod
    def _set_owner(self, path, owner=None, group=None):
        """Change the owner of file.

        :param path: path.
        :param owner: Optional, new owner for file.
        :param group: Optional, new group for file.
        At least one of `owner` and `group` must be specified.
        """
        raise NotImplementedError

    def set_owner(self, path, owner=None, group=None):
        rpath = self.resolve_path(path)
        if not self.exists(rpath):
            self._set_owner(rpath, owner=owner, group=group)
        else:
            raise FileSystemError("%r does not exist.", rpath)

    @abc.abstractmethod
    def _set_permission(self, path, permission):
        """Change the permissions of file.

        :param path: path.
        :param permission: New octal permissions string of file.
        """
        raise NotImplementedError

    def set_permission(self, path, permission=None):
        rpath = self.resolve_path(path)
        if not self.exists(rpath):
            self._set_permission(rpath, permission)
        else:
            raise FileSystemError("%r does not exist.", rpath)

    @abc.abstractmethod
    def _mkdir(self, path, permission=None):
        """Create a remote directory, recursively if necessary.

        :param path: Remote path. Intermediate directories will be created
          appropriately.
        :param permission: Octal permission to set on the newly created directory.
          These permissions will only be set on directories that do not already
          exist.
        """
        raise NotImplementedError

    def mkdir(self, path, permission=None):
        rpath = self.resolve_path(path)
        if not self.exists(rpath):
            self._mkdir(rpath, permission)
        else:
            raise FileSystemError("%r does not exist.", rpath)

    def walk(self, path, depth=0, status=False):
        """Depth-first walk of remote filesystem.

        :param path: Starting path. If the path doesn't exist, an
          :class:`FileSystemError` will be raised. If it points to a file, the returned
          generator will be empty.
        :param depth: Maximum depth to explore. `0` for no limit.
        :param status: Also return each file or folder's corresponding FileStatus_.
        This method returns a generator yielding tuples `(path, dirs, files)`
        where `path` is the absolute path to the current directory, `dirs` is the
        list of directory names it contains, and `files` is the list of file names
        it contains.
        """
        _logger.debug("Walking %r (depth %r).", path, depth)

        def _walk(dir_path, dir_status, depth):
            """Recursion helper."""
            infos = self.list(dir_path, status=True)
            dir_infos = [info for info in infos if info[1]["type"] == "DIRECTORY"]
            file_infos = [info for info in infos if info[1]["type"] == "FILE"]
            if status:
                yield ((self.format_path(dir_path), dir_status), dir_infos, file_infos)
            else:
                yield (
                    self.format_path(dir_path),
                    [name for name, _ in dir_infos],
                    [name for name, _ in file_infos],
                )
            if depth != 1:
                for name, s in dir_infos:
                    path = psp.join(dir_path, name)
                    for infos in _walk(path, s, depth - 1):
                        yield infos

        rpath = self.resolve_path(path)  # Cache resolution.
        s = self.status(rpath)
        if s["type"] == "DIRECTORY":
            for infos in _walk(rpath, s, depth):
                yield infos

    def checksum(self, path, algorithm="md5"):
        """Get checksum for a file.

        :param path: Remote path.
        :param algorithm: The checksum algorithm name.
        """
        if algorithm == "md5":
            hash_func = hashlib.md5
        elif algorithm == "sha1":
            hash_func = hashlib.sha1
        elif algorithm == "sha224":
            hash_func = hashlib.sha224
        elif algorithm == "sha256":
            hash_func = hashlib.sha256
        elif algorithm == "sha384":
            hash_func = hashlib.sha384
        elif algorithm == "sha512":
            hash_func = hashlib.sha512
        else:
            raise FileSystemError("Unknown checksum algorithm '%s'" % algorithm)

        checksum = hash_func()
        with self.read(path, chunk_size=8192) as reader:
            for chunk in reader:
                checksum.update(chunk)
        return checksum.hexdigest()

    @abc.abstractmethod
    def _open(self, path, mode, buffer_size=0, encoding=None, **kwargs):
        raise NotImplementedError

    def open(self, path, mode, buffer_size=1024, encoding=None, **kwargs):
        """Access a file from the Filesystem.

        Parameters
        ----------
        path: string
            Path of file on S3
        mode: string
            One of 'r', 'w', 'a', 'rb', 'wb', or 'ab'. These have the same meaning
            as they do for the built-in `open` function.
        buffer_size: int
            Size of data buffer when reading/writing data to the filesystem.
            The purpose of the buffer is to reduce the number of io.
        encoding : str
            The encoding to use if opening the file in text mode. The platform's
            default text encoding is used if not given.
        kwargs: dict-like
            Additional parameters used for file system specific usage.
        :return: A file object, idially an implementation of `BufferedIOBase`
        """
        rpath = self.resolve_path(path)
        return self._open(
            path=rpath, mode=mode, buffer_size=buffer_size, encoding=encoding, **kwargs
        )

    @contextmanager
    def read(
        self,
        path,
        offset=0,
        buffer_size=1024,
        encoding=None,
        chunk_size=None,
        delimiter=None,
        **kwargs
    ):
        """Read a file from FileSystem.

        :param path: path.
        :param offset: Starting byte position.
        :param length: Number of bytes to be processed. `None` will read the entire
          file.
        :param buffer_size: Size of the buffer in bytes used for transferring the
          data.
        :param encoding: Encoding used to decode the request. By default the raw
          data is returned. This is mostly helpful in python 3, for example to
          deserialize JSON data (as the decoder expects unicode).
        :param chunk_size: If set to a positive number, the context manager will
          return a generator yielding every `chunk_size` bytes instead of a
          file-like object (unless `delimiter` is also set, see below).
        :param delimiter: If set, the context manager will return a generator
          yielding each time the delimiter is encountered. This parameter requires
          the `encoding` to be specified.
        if `chunk_size` and `delimiter` are None, a file like object is returned,
        otherwise return a generator.
        This method must be called using a `with` block:
        .. code-block:: python
          with client.read('foo') as reader:
            content = reader.read()
        This ensures that connections are always properly closed.
        """
        if delimiter:
            if not encoding:
                raise ValueError("Delimiter splitting requires an encoding.")
            if chunk_size:
                raise ValueError("Delimiter splitting incompatible with chunk size.")

        rpath = self.resolve_path(path)
        if self.status(rpath, strict=False) is None:
            raise FileSystemError("%r does not exist.", rpath)

        _logger.debug("Reading file %r.", path)
        file = self.open(
            rpath, mode="rb", buffer_size=buffer_size, encoding=encoding, **kwargs
        )

        if offset > 0:
            file.seek(offset)
        try:
            if not chunk_size and not delimiter:
                # return a file like object
                yield file
            else:
                # return a generator function
                if delimiter:
                    yield DelimitedFileReader(file, delimiter=delimiter)
                else:
                    yield ChunkFileReader(file, chunk_size=chunk_size)
        finally:
            file.close()
            _logger.debug("Closed response for reading file %r.", path)

    def write(
        self,
        path,
        data=None,
        overwrite=False,
        permission=None,
        buffer_size=1024,
        append=False,
        encoding=None,
        **kwargs
    ):
        """Create a file.

        :param path: Path where to create file. The necessary directories will
          be created appropriately.
        :param data: Contents of file to write. Can be a string, a generator or a
          file object. The last two options will allow streaming upload (i.e.
          without having to load the entire contents into memory). If `None`, this
          method will return a file-like object and should be called using a `with`
          block (see below for examples).
        :param overwrite: Overwrite any existing file or directory.
        :param permission: Octal permission to set on the newly created file.
          Leading zeros may be omitted.
        :param buffer_size: Size of upload buffer.
        :param append: Append to a file rather than create a new one.
        :param encoding: Encoding used to serialize data written.
        :return: The file object returned by `open` if data is `None` else
        write the data to the file and return `void`.
        Sample usages:
        .. code-block:: python
          from json import dump, dumps
          records = [
            {'name': 'foo', 'weight': 1},
            {'name': 'bar', 'weight': 2},
          ]
          # As a context manager:
          with client.write('data/records.jsonl', encoding='utf-8') as writer:
            dump(records, writer)
          # Or, passing in a generator directly:
          client.write('data/records.jsonl', data=dumps(records), encoding='utf-8')
        """
        rpath = self.resolve_path(path)
        status = self.status(rpath, strict=False)
        if append:
            if overwrite:
                raise ValueError("Cannot both overwrite and append.")
            if permission:
                raise ValueError("Cannot change file properties while appending.")

            if status is not None and status["type"] != "FILE":
                raise ValueError("Path %r is not a file.", rpath)
        else:
            if not overwrite:
                if status is not None:
                    raise ValueError("Path %r exists, missing `append`.", rpath)
            else:
                if status is not None and status["type"] != "FILE":
                    raise ValueError("Path %r is not a file.", rpath)

        _logger.debug("Writing to %r.", path)
        file = self.open(
            rpath,
            mode="ab" if append else "wb",
            buffer_size=buffer_size,
            encoding=encoding,
            **kwargs
        )
        if data is None:
            return file
        else:
            with file:
                if isinstance(data, types.GeneratorType) or isinstance(
                    data, FileReader
                ):
                    for chunk in data:
                        file.write(chunk)
                else:
                    file.write(data)



import pytest

from tanit.common.model.execution_type import ExecutionType
from tanit.common.model.job import Job
from tanit.thrift.master.service.ttypes import JobInitializationException

from ..unit.utils.tutils import wait_until


@pytest.mark.usefixtures("master_server", "worker_server")
class TestServiceIntegration:
    def test_server_up(self, user_client):
        assert len(user_client.list_jobs()) >= 0

    def test_register_worker(self, worker_client):
        assert len(worker_client.list_workers()) == 1

    def test_submit_job(self, user_client):
        jid = user_client.submit_job(Job(ExecutionType.MOCK, {"num_tasks": "10"}))
        # wait for the job to finish
        assert wait_until(
            lambda i: user_client.job_status(i).state == "FINISHED", 20, 0.5, jid
        )

    def test_submit_job_failure(self, user_client):
        with pytest.raises(JobInitializationException):
            user_client.submit_job(Job(ExecutionType.MOCK, {"num_tasks": "not_an_int"}))

    def test_inexistent_job(self, user_client):
        assert user_client.job_status("some_inexistent_job") is None

    def test_job_failure(self, user_client):
        jid = user_client.submit_job(
            Job(ExecutionType.MOCK, {"num_tasks": "10", "num_failures": "2"})
        )
        # wait for the job to finish
        assert wait_until(
            lambda i: user_client.job_status(i).state == "FAILED", 20, 0.5, jid
        )

    def test_list_jobs(self, user_client):
        assert len(user_client.list_jobs()) >= 1



import pytest

from tanit.common.model.execution_type import ExecutionType
from tanit.common.model.job import Job
from tanit.master.core.execution.execution_job import (
    IllegalStateTransitionException,  # NOQA
)
from tanit.master.core.execution.execution_state import ExecutionState  # NOQA
from tanit.master.core.execution.job_factory import JobFactory  # NOQA

job_factory = JobFactory()


def mock_job_exec(num_tasks):
    job = job_factory.create_job(Job(ExecutionType.MOCK, {"num_tasks": str(num_tasks)}))
    job.setup()
    return job


class TestExecutionState:
    def test_initial_state(self):
        job = mock_job_exec(2)
        assert job.state == ExecutionState.SUBMITTED
        for task in job.get_tasks():
            assert task.state == ExecutionState.SUBMITTED

    def test_schedule_state_transition(self):
        job = mock_job_exec(2)

        job.get_tasks()[0].on_schedule()
        assert job.state == ExecutionState.SCHEDULED

        for task in job.get_tasks()[1:]:
            task.on_schedule()

        assert job.state == ExecutionState.SCHEDULED

    def test_dispatch_state(self):
        job = mock_job_exec(2)

        job.get_tasks()[0].on_schedule()
        job.get_tasks()[0].on_dispatch()
        assert job.state == ExecutionState.DISPATCHED

        for task in job.get_tasks()[1:]:
            task.on_schedule()
        assert job.state == ExecutionState.DISPATCHED

        for task in job.get_tasks()[1:]:
            task.on_dispatch()
        assert job.state == ExecutionState.DISPATCHED

    def test_running_state(self):
        job = mock_job_exec(2)

        job.get_tasks()[0].on_schedule()
        job.get_tasks()[0].on_dispatch()
        job.get_tasks()[0].on_start()
        assert job.state == ExecutionState.RUNNING

        for task in job.get_tasks()[1:]:
            task.on_schedule()
        assert job.state == ExecutionState.RUNNING

        for task in job.get_tasks()[1:]:
            task.on_dispatch()
        assert job.state == ExecutionState.RUNNING

        for task in job.get_tasks()[1:]:
            task.on_start()
        assert job.state == ExecutionState.RUNNING

    def test_running_state_2(self):
        job = mock_job_exec(2)

        job.get_tasks()[0].on_schedule()
        with pytest.raises(IllegalStateTransitionException):
            job.get_tasks()[0].on_start()

    def test_finish_state(self):
        job = mock_job_exec(2)

        job.get_tasks()[0].on_schedule()
        job.get_tasks()[0].on_dispatch()
        job.get_tasks()[0].on_start()
        job.get_tasks()[0].on_finish()
        assert job.state == ExecutionState.RUNNING

        for task in job.get_tasks()[1:]:
            task.on_schedule()
        assert job.state == ExecutionState.RUNNING

        for task in job.get_tasks()[1:]:
            task.on_dispatch()
        assert job.state == ExecutionState.RUNNING

        for task in job.get_tasks()[1:]:
            task.on_start()
        assert job.state == ExecutionState.RUNNING

        for task in job.get_tasks()[1:]:
            task.on_finish()
        assert job.state == ExecutionState.FINISHED

    def test_finish_state_2(self):
        job = mock_job_exec(2)

        job.get_tasks()[0].on_schedule()
        job.get_tasks()[0].on_dispatch()
        with pytest.raises(IllegalStateTransitionException):
            job.get_tasks()[0].on_finish()

    def test_fail_state_1(self):
        job = mock_job_exec(2)
        job.get_tasks()[0].on_schedule()
        job.get_tasks()[0].on_dispatch()
        job.get_tasks()[0].on_start()
        job.get_tasks()[0].on_fail()
        assert job.state == ExecutionState.FAILED

        for task in job.get_tasks()[1:]:
            task.on_schedule()
            task.on_dispatch()
            task.on_start()
            task.on_finish()
        assert job.state == ExecutionState.FAILED

    def test_fail_state_2(self):

        job = mock_job_exec(2)
        job.get_tasks()[0].on_schedule()
        job.get_tasks()[0].on_dispatch()
        job.get_tasks()[0].on_start()
        job.get_tasks()[0].on_finish()
        assert job.state == ExecutionState.RUNNING

        for task in job.get_tasks()[1:]:
            task.on_schedule()
            task.on_dispatch()
            task.on_start()
            task.on_fail()
        assert job.state == ExecutionState.FAILED

    def test_state_reset(self):
        job = mock_job_exec(2)
        job.get_tasks()[0].on_schedule()
        job.get_tasks()[0].on_dispatch()
        job.get_tasks()[0].on_start()
        job.get_tasks()[0].on_fail()
        assert job.state == ExecutionState.FAILED

        for task in job.get_tasks()[1:]:
            task.on_schedule()
            task.on_dispatch()
            task.on_start()
            task.on_finish()
        assert job.state == ExecutionState.FAILED

        job.get_tasks()[0].on_reset()
        assert job.state == ExecutionState.RUNNING

        job.get_tasks()[0].on_schedule()
        job.get_tasks()[0].on_dispatch()
        job.get_tasks()[0].on_start()
        assert job.state == ExecutionState.RUNNING

        job.get_tasks()[0].on_finish()
        assert job.state == ExecutionState.FINISHED



from ..common.core.exception import TanitError


class FileReader:
    pass


class ChunkFileReader(FileReader):
    def __init__(self, file, chunk_size=1024):
        self.file = file
        self.chunk_size = chunk_size

    def __iter__(self):
        return self

    def __next__(self):
        return self.next()

    def next(self):
        while True:
            data = self.file.read(self.chunk_size)
            if not data:
                raise StopIteration
            return data


class DelimitedFileReader(FileReader):
    def __init__(self, file, delimiter="\n"):
        self.file = file
        self.delimiter = delimiter
        self.end = False
        self.data = []
        self.buffer = ""

    def __iter__(self):
        return self

    def __next__(self):
        return self.next()

    def next(self):
        # TODO
        pass


class FileSystemError(TanitError):
    pass



class Worker(object):
    def __init__(self, wid, address, port):
        self.wid = wid
        self.address = address
        self.port = port

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            if self.wid == other.wid:
                if self.address == other.address:
                    if self.port == other.port:
                        return True
                    else:
                        return False
                else:
                    return False
            else:
                return False
        else:
            return False

    def __str__(self):
        return "Worker { id: %s, address: %s, port: %s}" % (
            self.wid,
            self.address,
            self.port,
        )


class WorkerStatus(object):
    def __init__(self, wid, running, pending, available):
        self.wid = wid
        self.running = running
        self.pending = pending
        self.available = available



import pytest

from tanit.master.standalone.master import StandaloneMaster


@pytest.fixture
def master():
    master = StandaloneMaster()
    master.start()
    yield master
    master.stop()


class TestStandaloneMaster:
    def test_master_setup(self, master):
        assert len(master.list_workers()) == 1



#
# Autogenerated by Thrift Compiler (0.13.0)
#
# DO NOT EDIT UNLESS YOU ARE SURE THAT YOU KNOW WHAT YOU ARE DOING
#
#  options string: py
#

from thrift.Thrift import TType, TMessageType, TFrozenDict, TException, TApplicationException
from thrift.protocol.TProtocol import TProtocolException
from thrift.TRecursive import fix_spec

import sys
import logging
from .ttypes import *
from thrift.Thrift import TProcessor
from thrift.transport import TTransport
all_structs = []


class Iface(object):
    def list_workers(self):
        pass

    def register_worker(self, worker):
        """
        Parameters:
         - worker

        """
        pass

    def unregister_worker(self, worker):
        """
        Parameters:
         - worker

        """
        pass

    def register_heartbeat(self, worker):
        """
        Parameters:
         - worker

        """
        pass

    def register_filesystem(self, filesystem):
        """
        Parameters:
         - filesystem

        """
        pass

    def task_start(self, tid):
        """
        Parameters:
         - tid

        """
        pass

    def task_success(self, tid):
        """
        Parameters:
         - tid

        """
        pass

    def task_failure(self, tid):
        """
        Parameters:
         - tid

        """
        pass


class Client(Iface):
    def __init__(self, iprot, oprot=None):
        self._iprot = self._oprot = iprot
        if oprot is not None:
            self._oprot = oprot
        self._seqid = 0

    def list_workers(self):
        self.send_list_workers()
        return self.recv_list_workers()

    def send_list_workers(self):
        self._oprot.writeMessageBegin('list_workers', TMessageType.CALL, self._seqid)
        args = list_workers_args()
        args.write(self._oprot)
        self._oprot.writeMessageEnd()
        self._oprot.trans.flush()

    def recv_list_workers(self):
        iprot = self._iprot
        (fname, mtype, rseqid) = iprot.readMessageBegin()
        if mtype == TMessageType.EXCEPTION:
            x = TApplicationException()
            x.read(iprot)
            iprot.readMessageEnd()
            raise x
        result = list_workers_result()
        result.read(iprot)
        iprot.readMessageEnd()
        if result.success is not None:
            return result.success
        raise TApplicationException(TApplicationException.MISSING_RESULT, "list_workers failed: unknown result")

    def register_worker(self, worker):
        """
        Parameters:
         - worker

        """
        self.send_register_worker(worker)
        self.recv_register_worker()

    def send_register_worker(self, worker):
        self._oprot.writeMessageBegin('register_worker', TMessageType.CALL, self._seqid)
        args = register_worker_args()
        args.worker = worker
        args.write(self._oprot)
        self._oprot.writeMessageEnd()
        self._oprot.trans.flush()

    def recv_register_worker(self):
        iprot = self._iprot
        (fname, mtype, rseqid) = iprot.readMessageBegin()
        if mtype == TMessageType.EXCEPTION:
            x = TApplicationException()
            x.read(iprot)
            iprot.readMessageEnd()
            raise x
        result = register_worker_result()
        result.read(iprot)
        iprot.readMessageEnd()
        return

    def unregister_worker(self, worker):
        """
        Parameters:
         - worker

        """
        self.send_unregister_worker(worker)
        self.recv_unregister_worker()

    def send_unregister_worker(self, worker):
        self._oprot.writeMessageBegin('unregister_worker', TMessageType.CALL, self._seqid)
        args = unregister_worker_args()
        args.worker = worker
        args.write(self._oprot)
        self._oprot.writeMessageEnd()
        self._oprot.trans.flush()

    def recv_unregister_worker(self):
        iprot = self._iprot
        (fname, mtype, rseqid) = iprot.readMessageBegin()
        if mtype == TMessageType.EXCEPTION:
            x = TApplicationException()
            x.read(iprot)
            iprot.readMessageEnd()
            raise x
        result = unregister_worker_result()
        result.read(iprot)
        iprot.readMessageEnd()
        return

    def register_heartbeat(self, worker):
        """
        Parameters:
         - worker

        """
        self.send_register_heartbeat(worker)
        self.recv_register_heartbeat()

    def send_register_heartbeat(self, worker):
        self._oprot.writeMessageBegin('register_heartbeat', TMessageType.CALL, self._seqid)
        args = register_heartbeat_args()
        args.worker = worker
        args.write(self._oprot)
        self._oprot.writeMessageEnd()
        self._oprot.trans.flush()

    def recv_register_heartbeat(self):
        iprot = self._iprot
        (fname, mtype, rseqid) = iprot.readMessageBegin()
        if mtype == TMessageType.EXCEPTION:
            x = TApplicationException()
            x.read(iprot)
            iprot.readMessageEnd()
            raise x
        result = register_heartbeat_result()
        result.read(iprot)
        iprot.readMessageEnd()
        return

    def register_filesystem(self, filesystem):
        """
        Parameters:
         - filesystem

        """
        self.send_register_filesystem(filesystem)
        self.recv_register_filesystem()

    def send_register_filesystem(self, filesystem):
        self._oprot.writeMessageBegin('register_filesystem', TMessageType.CALL, self._seqid)
        args = register_filesystem_args()
        args.filesystem = filesystem
        args.write(self._oprot)
        self._oprot.writeMessageEnd()
        self._oprot.trans.flush()

    def recv_register_filesystem(self):
        iprot = self._iprot
        (fname, mtype, rseqid) = iprot.readMessageBegin()
        if mtype == TMessageType.EXCEPTION:
            x = TApplicationException()
            x.read(iprot)
            iprot.readMessageEnd()
            raise x
        result = register_filesystem_result()
        result.read(iprot)
        iprot.readMessageEnd()
        return

    def task_start(self, tid):
        """
        Parameters:
         - tid

        """
        self.send_task_start(tid)
        self.recv_task_start()

    def send_task_start(self, tid):
        self._oprot.writeMessageBegin('task_start', TMessageType.CALL, self._seqid)
        args = task_start_args()
        args.tid = tid
        args.write(self._oprot)
        self._oprot.writeMessageEnd()
        self._oprot.trans.flush()

    def recv_task_start(self):
        iprot = self._iprot
        (fname, mtype, rseqid) = iprot.readMessageBegin()
        if mtype == TMessageType.EXCEPTION:
            x = TApplicationException()
            x.read(iprot)
            iprot.readMessageEnd()
            raise x
        result = task_start_result()
        result.read(iprot)
        iprot.readMessageEnd()
        return

    def task_success(self, tid):
        """
        Parameters:
         - tid

        """
        self.send_task_success(tid)
        self.recv_task_success()

    def send_task_success(self, tid):
        self._oprot.writeMessageBegin('task_success', TMessageType.CALL, self._seqid)
        args = task_success_args()
        args.tid = tid
        args.write(self._oprot)
        self._oprot.writeMessageEnd()
        self._oprot.trans.flush()

    def recv_task_success(self):
        iprot = self._iprot
        (fname, mtype, rseqid) = iprot.readMessageBegin()
        if mtype == TMessageType.EXCEPTION:
            x = TApplicationException()
            x.read(iprot)
            iprot.readMessageEnd()
            raise x
        result = task_success_result()
        result.read(iprot)
        iprot.readMessageEnd()
        return

    def task_failure(self, tid):
        """
        Parameters:
         - tid

        """
        self.send_task_failure(tid)
        self.recv_task_failure()

    def send_task_failure(self, tid):
        self._oprot.writeMessageBegin('task_failure', TMessageType.CALL, self._seqid)
        args = task_failure_args()
        args.tid = tid
        args.write(self._oprot)
        self._oprot.writeMessageEnd()
        self._oprot.trans.flush()

    def recv_task_failure(self):
        iprot = self._iprot
        (fname, mtype, rseqid) = iprot.readMessageBegin()
        if mtype == TMessageType.EXCEPTION:
            x = TApplicationException()
            x.read(iprot)
            iprot.readMessageEnd()
            raise x
        result = task_failure_result()
        result.read(iprot)
        iprot.readMessageEnd()
        return


class Processor(Iface, TProcessor):
    def __init__(self, handler):
        self._handler = handler
        self._processMap = {}
        self._processMap["list_workers"] = Processor.process_list_workers
        self._processMap["register_worker"] = Processor.process_register_worker
        self._processMap["unregister_worker"] = Processor.process_unregister_worker
        self._processMap["register_heartbeat"] = Processor.process_register_heartbeat
        self._processMap["register_filesystem"] = Processor.process_register_filesystem
        self._processMap["task_start"] = Processor.process_task_start
        self._processMap["task_success"] = Processor.process_task_success
        self._processMap["task_failure"] = Processor.process_task_failure
        self._on_message_begin = None

    def on_message_begin(self, func):
        self._on_message_begin = func

    def process(self, iprot, oprot):
        (name, type, seqid) = iprot.readMessageBegin()
        if self._on_message_begin:
            self._on_message_begin(name, type, seqid)
        if name not in self._processMap:
            iprot.skip(TType.STRUCT)
            iprot.readMessageEnd()
            x = TApplicationException(TApplicationException.UNKNOWN_METHOD, 'Unknown function %s' % (name))
            oprot.writeMessageBegin(name, TMessageType.EXCEPTION, seqid)
            x.write(oprot)
            oprot.writeMessageEnd()
            oprot.trans.flush()
            return
        else:
            self._processMap[name](self, seqid, iprot, oprot)
        return True

    def process_list_workers(self, seqid, iprot, oprot):
        args = list_workers_args()
        args.read(iprot)
        iprot.readMessageEnd()
        result = list_workers_result()
        try:
            result.success = self._handler.list_workers()
            msg_type = TMessageType.REPLY
        except TTransport.TTransportException:
            raise
        except TApplicationException as ex:
            logging.exception('TApplication exception in handler')
            msg_type = TMessageType.EXCEPTION
            result = ex
        except Exception:
            logging.exception('Unexpected exception in handler')
            msg_type = TMessageType.EXCEPTION
            result = TApplicationException(TApplicationException.INTERNAL_ERROR, 'Internal error')
        oprot.writeMessageBegin("list_workers", msg_type, seqid)
        result.write(oprot)
        oprot.writeMessageEnd()
        oprot.trans.flush()

    def process_register_worker(self, seqid, iprot, oprot):
        args = register_worker_args()
        args.read(iprot)
        iprot.readMessageEnd()
        result = register_worker_result()
        try:
            self._handler.register_worker(args.worker)
            msg_type = TMessageType.REPLY
        except TTransport.TTransportException:
            raise
        except TApplicationException as ex:
            logging.exception('TApplication exception in handler')
            msg_type = TMessageType.EXCEPTION
            result = ex
        except Exception:
            logging.exception('Unexpected exception in handler')
            msg_type = TMessageType.EXCEPTION
            result = TApplicationException(TApplicationException.INTERNAL_ERROR, 'Internal error')
        oprot.writeMessageBegin("register_worker", msg_type, seqid)
        result.write(oprot)
        oprot.writeMessageEnd()
        oprot.trans.flush()

    def process_unregister_worker(self, seqid, iprot, oprot):
        args = unregister_worker_args()
        args.read(iprot)
        iprot.readMessageEnd()
        result = unregister_worker_result()
        try:
            self._handler.unregister_worker(args.worker)
            msg_type = TMessageType.REPLY
        except TTransport.TTransportException:
            raise
        except TApplicationException as ex:
            logging.exception('TApplication exception in handler')
            msg_type = TMessageType.EXCEPTION
            result = ex
        except Exception:
            logging.exception('Unexpected exception in handler')
            msg_type = TMessageType.EXCEPTION
            result = TApplicationException(TApplicationException.INTERNAL_ERROR, 'Internal error')
        oprot.writeMessageBegin("unregister_worker", msg_type, seqid)
        result.write(oprot)
        oprot.writeMessageEnd()
        oprot.trans.flush()

    def process_register_heartbeat(self, seqid, iprot, oprot):
        args = register_heartbeat_args()
        args.read(iprot)
        iprot.readMessageEnd()
        result = register_heartbeat_result()
        try:
            self._handler.register_heartbeat(args.worker)
            msg_type = TMessageType.REPLY
        except TTransport.TTransportException:
            raise
        except TApplicationException as ex:
            logging.exception('TApplication exception in handler')
            msg_type = TMessageType.EXCEPTION
            result = ex
        except Exception:
            logging.exception('Unexpected exception in handler')
            msg_type = TMessageType.EXCEPTION
            result = TApplicationException(TApplicationException.INTERNAL_ERROR, 'Internal error')
        oprot.writeMessageBegin("register_heartbeat", msg_type, seqid)
        result.write(oprot)
        oprot.writeMessageEnd()
        oprot.trans.flush()

    def process_register_filesystem(self, seqid, iprot, oprot):
        args = register_filesystem_args()
        args.read(iprot)
        iprot.readMessageEnd()
        result = register_filesystem_result()
        try:
            self._handler.register_filesystem(args.filesystem)
            msg_type = TMessageType.REPLY
        except TTransport.TTransportException:
            raise
        except TApplicationException as ex:
            logging.exception('TApplication exception in handler')
            msg_type = TMessageType.EXCEPTION
            result = ex
        except Exception:
            logging.exception('Unexpected exception in handler')
            msg_type = TMessageType.EXCEPTION
            result = TApplicationException(TApplicationException.INTERNAL_ERROR, 'Internal error')
        oprot.writeMessageBegin("register_filesystem", msg_type, seqid)
        result.write(oprot)
        oprot.writeMessageEnd()
        oprot.trans.flush()

    def process_task_start(self, seqid, iprot, oprot):
        args = task_start_args()
        args.read(iprot)
        iprot.readMessageEnd()
        result = task_start_result()
        try:
            self._handler.task_start(args.tid)
            msg_type = TMessageType.REPLY
        except TTransport.TTransportException:
            raise
        except TApplicationException as ex:
            logging.exception('TApplication exception in handler')
            msg_type = TMessageType.EXCEPTION
            result = ex
        except Exception:
            logging.exception('Unexpected exception in handler')
            msg_type = TMessageType.EXCEPTION
            result = TApplicationException(TApplicationException.INTERNAL_ERROR, 'Internal error')
        oprot.writeMessageBegin("task_start", msg_type, seqid)
        result.write(oprot)
        oprot.writeMessageEnd()
        oprot.trans.flush()

    def process_task_success(self, seqid, iprot, oprot):
        args = task_success_args()
        args.read(iprot)
        iprot.readMessageEnd()
        result = task_success_result()
        try:
            self._handler.task_success(args.tid)
            msg_type = TMessageType.REPLY
        except TTransport.TTransportException:
            raise
        except TApplicationException as ex:
            logging.exception('TApplication exception in handler')
            msg_type = TMessageType.EXCEPTION
            result = ex
        except Exception:
            logging.exception('Unexpected exception in handler')
            msg_type = TMessageType.EXCEPTION
            result = TApplicationException(TApplicationException.INTERNAL_ERROR, 'Internal error')
        oprot.writeMessageBegin("task_success", msg_type, seqid)
        result.write(oprot)
        oprot.writeMessageEnd()
        oprot.trans.flush()

    def process_task_failure(self, seqid, iprot, oprot):
        args = task_failure_args()
        args.read(iprot)
        iprot.readMessageEnd()
        result = task_failure_result()
        try:
            self._handler.task_failure(args.tid)
            msg_type = TMessageType.REPLY
        except TTransport.TTransportException:
            raise
        except TApplicationException as ex:
            logging.exception('TApplication exception in handler')
            msg_type = TMessageType.EXCEPTION
            result = ex
        except Exception:
            logging.exception('Unexpected exception in handler')
            msg_type = TMessageType.EXCEPTION
            result = TApplicationException(TApplicationException.INTERNAL_ERROR, 'Internal error')
        oprot.writeMessageBegin("task_failure", msg_type, seqid)
        result.write(oprot)
        oprot.writeMessageEnd()
        oprot.trans.flush()

# HELPER FUNCTIONS AND STRUCTURES


class list_workers_args(object):


    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('list_workers_args')
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(list_workers_args)
list_workers_args.thrift_spec = (
)


class list_workers_result(object):
    """
    Attributes:
     - success

    """


    def __init__(self, success=None,):
        self.success = success

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 0:
                if ftype == TType.LIST:
                    self.success = []
                    (_etype28, _size25) = iprot.readListBegin()
                    for _i29 in range(_size25):
                        _elem30 = Worker()
                        _elem30.read(iprot)
                        self.success.append(_elem30)
                    iprot.readListEnd()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('list_workers_result')
        if self.success is not None:
            oprot.writeFieldBegin('success', TType.LIST, 0)
            oprot.writeListBegin(TType.STRUCT, len(self.success))
            for iter31 in self.success:
                iter31.write(oprot)
            oprot.writeListEnd()
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(list_workers_result)
list_workers_result.thrift_spec = (
    (0, TType.LIST, 'success', (TType.STRUCT, [Worker, None], False), None, ),  # 0
)


class register_worker_args(object):
    """
    Attributes:
     - worker

    """


    def __init__(self, worker=None,):
        self.worker = worker

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRUCT:
                    self.worker = Worker()
                    self.worker.read(iprot)
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('register_worker_args')
        if self.worker is not None:
            oprot.writeFieldBegin('worker', TType.STRUCT, 1)
            self.worker.write(oprot)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(register_worker_args)
register_worker_args.thrift_spec = (
    None,  # 0
    (1, TType.STRUCT, 'worker', [Worker, None], None, ),  # 1
)


class register_worker_result(object):


    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('register_worker_result')
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(register_worker_result)
register_worker_result.thrift_spec = (
)


class unregister_worker_args(object):
    """
    Attributes:
     - worker

    """


    def __init__(self, worker=None,):
        self.worker = worker

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRUCT:
                    self.worker = Worker()
                    self.worker.read(iprot)
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('unregister_worker_args')
        if self.worker is not None:
            oprot.writeFieldBegin('worker', TType.STRUCT, 1)
            self.worker.write(oprot)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(unregister_worker_args)
unregister_worker_args.thrift_spec = (
    None,  # 0
    (1, TType.STRUCT, 'worker', [Worker, None], None, ),  # 1
)


class unregister_worker_result(object):


    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('unregister_worker_result')
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(unregister_worker_result)
unregister_worker_result.thrift_spec = (
)


class register_heartbeat_args(object):
    """
    Attributes:
     - worker

    """


    def __init__(self, worker=None,):
        self.worker = worker

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRUCT:
                    self.worker = Worker()
                    self.worker.read(iprot)
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('register_heartbeat_args')
        if self.worker is not None:
            oprot.writeFieldBegin('worker', TType.STRUCT, 1)
            self.worker.write(oprot)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(register_heartbeat_args)
register_heartbeat_args.thrift_spec = (
    None,  # 0
    (1, TType.STRUCT, 'worker', [Worker, None], None, ),  # 1
)


class register_heartbeat_result(object):


    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('register_heartbeat_result')
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(register_heartbeat_result)
register_heartbeat_result.thrift_spec = (
)


class register_filesystem_args(object):
    """
    Attributes:
     - filesystem

    """


    def __init__(self, filesystem=None,):
        self.filesystem = filesystem

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRUCT:
                    self.filesystem = FileSystem()
                    self.filesystem.read(iprot)
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('register_filesystem_args')
        if self.filesystem is not None:
            oprot.writeFieldBegin('filesystem', TType.STRUCT, 1)
            self.filesystem.write(oprot)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(register_filesystem_args)
register_filesystem_args.thrift_spec = (
    None,  # 0
    (1, TType.STRUCT, 'filesystem', [FileSystem, None], None, ),  # 1
)


class register_filesystem_result(object):


    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('register_filesystem_result')
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(register_filesystem_result)
register_filesystem_result.thrift_spec = (
)


class task_start_args(object):
    """
    Attributes:
     - tid

    """


    def __init__(self, tid=None,):
        self.tid = tid

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRING:
                    self.tid = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('task_start_args')
        if self.tid is not None:
            oprot.writeFieldBegin('tid', TType.STRING, 1)
            oprot.writeString(self.tid.encode('utf-8') if sys.version_info[0] == 2 else self.tid)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(task_start_args)
task_start_args.thrift_spec = (
    None,  # 0
    (1, TType.STRING, 'tid', 'UTF8', None, ),  # 1
)


class task_start_result(object):


    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('task_start_result')
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(task_start_result)
task_start_result.thrift_spec = (
)


class task_success_args(object):
    """
    Attributes:
     - tid

    """


    def __init__(self, tid=None,):
        self.tid = tid

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRING:
                    self.tid = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('task_success_args')
        if self.tid is not None:
            oprot.writeFieldBegin('tid', TType.STRING, 1)
            oprot.writeString(self.tid.encode('utf-8') if sys.version_info[0] == 2 else self.tid)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(task_success_args)
task_success_args.thrift_spec = (
    None,  # 0
    (1, TType.STRING, 'tid', 'UTF8', None, ),  # 1
)


class task_success_result(object):


    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('task_success_result')
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(task_success_result)
task_success_result.thrift_spec = (
)


class task_failure_args(object):
    """
    Attributes:
     - tid

    """


    def __init__(self, tid=None,):
        self.tid = tid

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRING:
                    self.tid = iprot.readString().decode('utf-8') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('task_failure_args')
        if self.tid is not None:
            oprot.writeFieldBegin('tid', TType.STRING, 1)
            oprot.writeString(self.tid.encode('utf-8') if sys.version_info[0] == 2 else self.tid)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(task_failure_args)
task_failure_args.thrift_spec = (
    None,  # 0
    (1, TType.STRING, 'tid', 'UTF8', None, ),  # 1
)


class task_failure_result(object):


    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('task_failure_result')
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(task_failure_result)
task_failure_result.thrift_spec = (
)
fix_spec(all_structs)
del all_structs




import pytest
from six.moves.queue import Queue

from tanit.common.model.execution_type import ExecutionType
from tanit.common.model.job import Job
from tanit.master.core.execution.job_factory import JobFactory
from tanit.master.core.scheduler import SimpleScheduler

from ...utils.tutils import wait_until

job_factory = JobFactory()


def mock_job_exec(num_tasks):
    job = job_factory.create_job(Job(ExecutionType.MOCK, {"num_tasks": str(num_tasks)}))
    job.setup()
    return job


def _verify_queue_size(cqueue, size):
    return cqueue.qsize() == size


@pytest.fixture
def simple_scheduler():
    simple_scheduler = SimpleScheduler(Queue(), Queue(), None)
    simple_scheduler.start()

    yield simple_scheduler

    simple_scheduler.stop()


class TestSimpleScheduler:
    def test_schedule(self, simple_scheduler):

        for task in mock_job_exec(2).get_tasks():
            simple_scheduler.lqueue.put(task)

        assert wait_until(_verify_queue_size, 10, 0.5, simple_scheduler.lqueue, 0)
        assert wait_until(_verify_queue_size, 10, 0.5, simple_scheduler.cqueue, 2)

    def test_scheduler_stop(self, simple_scheduler):
        simple_scheduler.stop()

        for task in mock_job_exec(2).get_tasks():
            simple_scheduler.lqueue.put(task)

        assert wait_until(_verify_queue_size, 10, 0.5, simple_scheduler.lqueue, 2)
        assert wait_until(_verify_queue_size, 10, 0.5, simple_scheduler.cqueue, 0)

    def test_scheduler_callback(self, simple_scheduler):
        def callback(tid):
            callback_received.append(tid)

        callback_received = []
        simple_scheduler.callback = callback

        for task in mock_job_exec(2).get_tasks():
            simple_scheduler.lqueue.put(task)

        assert wait_until(lambda l, size: len(l) == size, 10, 0.5, callback_received, 2)



import pytest

from tanit.common.model.execution_type import ExecutionType
from tanit.common.model.job import Job
from tanit.common.model.worker import Worker
from tanit.master.core.execution.execution_job import JobInitializationException
from tanit.master.core.execution.execution_state import ExecutionState
from tanit.master.core.master import Master

from ...utils.tutils import wait_until


@pytest.fixture
def master():
    master = Master()
    master.start()
    yield master
    master.stop()


class TestMaster:
    def test_submit_job(self, master):
        master.register_worker(Worker("worker 1", None, None))
        job = master.submit_job(Job(ExecutionType.MOCK, {"num_tasks": "2"}))
        assert wait_until(
            lambda i: master.get_job(i).state == ExecutionState.FINISHED,
            20,
            0.5,
            job.jid,
        )

    def test_submit_bad_job_params(self, master):
        master.register_worker(Worker("worker 1", None, None))
        with pytest.raises(JobInitializationException):
            master.submit_job(Job(ExecutionType.MOCK, {"num_tasks": "not_an_int"}))

    def test_list_jobs(self, master):
        master.submit_job(Job(ExecutionType.MOCK, {"num_tasks": "2"}))
        master.register_worker(Worker("worker 1", None, None))
        assert len(master.list_jobs()) == 1

    def test_register_worker(self, master):
        master.register_worker(Worker("worker 1", None, None))
        assert len(master.list_workers()) == 1

    def test_register_heart_beat(self, master):
        master.register_worker(Worker("worker 1", None, None))
        master.register_heartbeat(Worker("worker 1", None, None))
        assert len(master.list_workers()) == 1

    def test_unregister_worker(self, master):
        master.register_worker(Worker("worker 1", None, None))
        assert len(master.list_workers()) == 1
        master.unregister_worker(Worker("worker 1", None, None))
        assert len(master.list_workers()) == 0
        master.register_worker(Worker("worker 2", None, None))
        assert len(master.list_workers()) == 1



class ExecutionState:
    SUBMITTED = 1
    SCHEDULED = 2
    DISPATCHED = 3
    RUNNING = 4
    FINISHED = 5
    FAILED = 6

    _VALUES_TO_NAMES = {
        1: "SUBMITTED",
        2: "SCHEDULED",
        3: "DISPATCHED",
        4: "RUNNING",
        5: "FINISHED",
        6: "FAILED",
    }

    _NAMES_TO_VALUES = {
        "SUBMITTED": 1,
        "SCHEDULED": 2,
        "DISPATCHED": 3,
        "RUNNING": 4,
        "FINISHED": 5,
        "FAILED": 6,
    }



import logging as lg

from six.moves.queue import Queue

from ..dispatcher import FairDispatcher
from ..scheduler import SimpleScheduler
from .job_factory import JobFactory

_logger = lg.getLogger(__name__)


class ExecutionManager(object):
    """
    The ExecutionManager monitor the state of execution of jobs and tasks.

    It interacts with the different components involved in
    the execution pipeline and ensure the execution state is properly updated
    and reflect the real progress.
    """

    def __init__(self, workers_manager, config=None):
        # jobs list
        self.jobs = []
        # job factory
        self.jobs_factory = JobFactory()
        # Lister queue
        lqueue = Queue()
        # Call queue
        cqueue = Queue()
        # scheduler
        self.scheduler = SimpleScheduler(lqueue, cqueue, self.task_schedule)
        # dispatcher
        self.dispatcher = FairDispatcher(cqueue, workers_manager, self.task_dispatch)

        self.configure(config)

    def configure(self, config):
        # should be configurable
        self.max_task_retries = 3

    def start(self):
        _logger.info("Stating Tanit master services.")
        self.dispatcher.start()
        self.scheduler.start()
        _logger.info("Tanit master services started.")

    def stop(self):
        _logger.info("Stopping Tanit master services.")
        self.scheduler.stop()
        self.dispatcher.stop()
        _logger.info("Tanit master services stopped.")

    def submit_job(self, job):
        job_exec = self.jobs_factory.create_job(job)
        job_exec.setup()

        _logger.info("Submitting job [ %s ] for execution.", job_exec.jid)
        self.jobs.append(job_exec)
        for task_exec in self.get_tasks(jid=job_exec.jid):
            self.scheduler.schedule(task_exec)

        _logger.info(
            "Submitted %s tasks for execution in job [ %s ].",
            len(self.get_tasks(jid=job_exec.jid)),
            job_exec.jid,
        )
        return job_exec

    def cancel_job(self, conf):
        pass

    def list_jobs(self):
        return self.jobs

    def get_job(self, jid):
        for job_exec in self.jobs:
            if job_exec.jid == jid:
                return job_exec
        return None

    def get_tasks(self, jid=None, states=[], worker=None):
        target_jobs = []
        if jid is not None:
            job = self.get_job(jid)
            if job is not None:
                target_jobs.append(job)
            else:
                return []
        else:
            target_jobs = self.jobs

        target_tasks = []
        for job in target_jobs:
            for task in job.get_tasks():
                valid = True
                if len(states) != 0:
                    if task.state not in states:
                        valid = False
                if worker is not None:
                    if task.worker != worker:
                        valid = False
                if valid:
                    target_tasks.append(task)
        return target_tasks

    def get_task(self, tid):
        for job_exec in self.jobs:
            task = job_exec.get_task(tid)
            if task is not None:
                return task
        return None

    def task_schedule(self, tid):
        task = self.get_task(tid)
        if task is not None:
            task.on_schedule()
        else:
            raise NoSuchTaskException("No such task [ %s ]", tid)

    def task_dispatch(self, tid, worker=None):
        task = self.get_task(tid)
        if task is not None:
            task.on_dispatch(worker)
        else:
            raise NoSuchTaskException("No such task [ %s ]", tid)

    def task_start(self, tid):
        task = self.get_task(tid)
        if task is not None:
            task.on_start()
        else:
            raise NoSuchTaskException("No such task [ %s ]", tid)

    def task_finish(self, tid):
        task = self.get_task(tid)
        if task is not None:
            task.on_finish()
        else:
            raise NoSuchTaskException("No such task [ %s ]", tid)

    def task_failure(self, tid):
        task = self.get_task(tid)
        if task is not None:
            _logger.info("Received task failure for %s" % tid)
            if task.attempts == self.max_task_retries:
                _logger.info("Failing task %s after %s attempts" % (tid, task.attempts))
                task.on_fail()
            else:
                _logger.info(
                    "Retrying task %s, attempt %s out of %s"
                    % (tid, task.attempts, self.max_task_retries)
                )
                task.on_retry()
                self.scheduler.schedule(task)
        else:
            raise NoSuchTaskException("No such task [ %s ]", tid)

    def task_reset(self, tid):
        task = self.get_task(tid)
        if task is not None:
            task.on_reset()
            self.scheduler.schedule(task)
        else:
            raise NoSuchTaskException("No such task [ %s ]", tid)


class NoSuchJobException(Exception):
    pass


class NoSuchTaskException(Exception):
    pass



import logging as lg
import multiprocessing

from six.moves.queue import Queue

from ...common.model.task import Task
from ...common.model.worker import WorkerStatus
from ...worker.core.execution.task_factory import TaskFactory
from ...worker.core.executor_factory import ExecutorFactory
from ...worker.core.executor_pool import ExecutorPool
from ..client.client import LocalClientFactory
from ..core.worker.worker import WorkerIFace

_logger = lg.getLogger(__name__)


class LocalWorker(WorkerIFace):
    def __init__(self, wid, master):
        super(LocalWorker, self).__init__(wid, None, None)
        self.lqueue = Queue()
        self.stopped = False
        self.task_factory = TaskFactory()

        client_factory = LocalClientFactory(master)

        self.executor = ExecutorPool(
            self.wid,
            ExecutorFactory(client_factory, self.lqueue, multiprocessing.cpu_count()),
            self.lqueue,
            multiprocessing.cpu_count(),
        )

    def submit(self, task):
        task_exec = self.task_factory.create_task(
            Task(task.tid, task.etype, task.params)
        )
        if not self.stopped:
            self.lqueue.put(task_exec)
        else:
            raise WorkerStoppedException(
                "Can not submit task [ %s ] to [ %s ] : worker stopped.",
                task_exec.tid,
                self.wid,
            )

    def register_filesystem(self, name, filesystem):
        # The file system is defined in the master already
        pass

    def status(self):
        return WorkerStatus(
            self.wid,
            self.executor.num_running(),
            self.executor.num_pending(),
            self.executor.num_available(),
        )

    def start(self):
        _logger.info("Starting tanit worker [%s].", self.wid)
        self.stopped = False
        self.executor.start()
        _logger.info("Tanit worker [%s] started.", self.wid)

    def stop(self):
        _logger.info("Stopping Tanit worker [ %s ].", self.wid)
        self.stopped = True
        self.executor.stop()
        _logger.info("Tanit worker [ %s ] stopped.", self.wid)


class WorkerStoppedException(Exception):
    """Raised when trying to submit a task to a stopped worker."""

    pass



import logging as lg
import os

from ...worker.filesystem.client import LocalFileSystemClient
from ..filesystem import IFileSystem

_logger = lg.getLogger(__name__)


class RemoteFileSystem(IFileSystem):
    def __init__(self, host, port):
        self.client = LocalFileSystemClient(host, port)
        self.client.start()

    def resolve_path(self, path):
        return os.path.normpath(os.path.abspath(path))

    def _list(self, path, status=False, glob=False):
        return self.client.ls(path, status, glob)

    def _status(self, path, strict=True):
        return self.client.status(path, strict)

    def _content(self, path, strict=True):
        return self.client.content(path, strict)

    def _delete(self, path, recursive=False):
        return self.client.rm(path, recursive)

    def copy(self, src_path, dst_path, recursive=True):
        return self.client.copy(src_path, dst_path)

    def _copy(self, src_path, dst_path):
        return self.client.copy(src_path, dst_path)

    def rename(self, src_path, dst_path):
        return self.client.rename(src_path, dst_path)

    def _set_owner(self, path, owner=None, group=None):
        return self.client.set_owner(path, owner, group)

    def _set_permission(self, path, permission):
        return self.client.set_permission(path, permission)

    def _mkdir(self, path, permission=None):
        return self.client.mkdir(path, permission)

    def _open(self, path, mode, buffer_size=0, encoding=None, **kwargs):
        return self.client.open(
            path,
            mode=mode,
            buffer_size=buffer_size,
            encoding=encoding,
        )



import logging as lg
import time
from threading import Thread

from six.moves.queue import Queue

from ...common.model.worker import WorkerStatus
from ...filesystem.filesystem_factory import FileSystemFactory
from ...master.client.client import ClientType
from ...master.client.client import ThriftClientFactory
from ..core.execution.task_factory import TaskFactory
from ..core.executor_factory import ExecutorFactory
from ..core.executor_pool import ExecutorPool
from ..filesystem.service import LocalFileSystemService

_logger = lg.getLogger(__name__)


class Worker(object):
    def __init__(self):
        self.lqueue = Queue()
        self.stopped = False

    def configure(self, config):
        self.address = config.worker_host
        self.port = config.worker_port

        self.wid = "tanit-worker-%s-%s" % (self.address, self.port)

        client_factory = ThriftClientFactory(config.master_host, config.master_port)
        self.master = client_factory.create_client(ClientType.WORKER_SERVICE)

        self.executor = ExecutorPool(
            self.wid,
            ExecutorFactory(client_factory, self.lqueue, config.executor_threads),
            self.lqueue,
            config.executor_threads,
        )

        self.task_factory = TaskFactory()

        self.reporter = WorkerHearbeatReporter(self)
        self.reporter.setDaemon(True)

        self.filesystem = LocalFileSystemService()

    def submit(self, task):
        task_exec = self.task_factory.create_task(task)
        if not self.stopped:
            self.lqueue.put(task_exec)
        else:
            raise WorkerStoppedException(
                "Can not submit task [ %s ] to [ %s ] : worker stopped.",
                task_exec.tid,
                self.wid,
            )

    def get_stats(self):
        return WorkerStatus(
            self.wid,
            self.executor.num_running(),
            self.executor.num_pending(),
            self.executor.num_available(),
        )

    def register_filesystem(self, name, filesystem):
        if self.stopped:
            raise WorkerStoppedException(
                "Can not register filesystem [ %s ] : worker server stopped.", name
            )

        _logger.info("Registering new filesystem [ %s ].", name)
        filesystem["name"] = name
        # register the worker as a filesystem
        FileSystemFactory.getInstance().register_filesystem(filesystem)
        _logger.info("Filesystem [ %s ] registered.", name)

    def start(self):
        _logger.info("Starting tanit worker [%s].", self.wid)
        self.stopped = False
        try:
            self.master.start()
        except Exception as e:
            _logger.error(
                "Could not connect to master on [%s:%s]", self.address, self.port
            )
            raise e

        self.executor.start()

        # register the worker
        self.master.register_worker(self.wid, self.address, self.port)

        # start the filesystem service
        self.filesystem.start()
        # register the file system
        if self.filesystem:
            # register the file system with the master
            self.master.register_filesystem(
                "local:%s" % self.address,
                {
                    "type": "local",
                    "address": str(self.address),
                    "port": str(self.filesystem.bind_port),
                },
            )

        self.reporter.start()

    def stop(self):
        _logger.info("Stopping Tanit worker [ %s ].", self.wid)
        self.stopped = True
        self.executor.stop()
        # unregister the worker
        try:
            self.master.unregister_worker(self.wid, self.address, self.port)
        except Exception:
            _logger.error("Could not unregister worker from master, exiting.")

        self.reporter.stop()
        self.reporter.join()

        self.master.stop()
        _logger.info("Tanit worker [ %s ] stopped.", self.wid)


class WorkerHearbeatReporter(Thread):
    # in seconds
    heartbeat_interval = 3

    def __init__(self, worker):
        super(WorkerHearbeatReporter, self).__init__()
        self.worker = worker
        self.client = worker.master
        self.stopped = False

    def stop(self):
        _logger.info("Stopping tanit worker hearbeat reporter.")
        self.stopped = True

    def run(self):
        _logger.info("Started tanit worker hearbeat reporter.")
        while not self.stopped:
            try:
                self.client.register_heartbeat(
                    self.worker.wid, self.worker.address, self.worker.port
                )
            except Exception:
                _logger.exception(
                    "Could not send heartbeat to master, is the server running !"
                )
            time.sleep(self.heartbeat_interval)
        _logger.info("Tanit worker hearbeat reporter stopped.")


class WorkerStoppedException(Exception):
    """Raised when trying to submit a task to a stopped worker."""

    pass



#!/usr/bin/env python

import logging as lg
import multiprocessing
import socket

from ...common.config.config import Config
from ...common.config.config import TanitConfigurationException

_logger = lg.getLogger(__name__)


class WorkerConfig(Config):
    default_bind_port = 9093
    default_bind_address = "0.0.0.0"
    default_thrift_threads = 25
    default_executor_threads = multiprocessing.cpu_count()

    def __init__(self, path=None):
        super(WorkerConfig, self).__init__(path)

    def load(self):
        if self.config.has_option("master", "worker-service_address"):
            master = self.config.get("master", "worker-service_address")
        else:
            raise TanitConfigurationException(
                "Missing master rpc worker service address from configuration"
            )

        if len(master.split(":")) != 2:
            raise TanitConfigurationException(
                "Master rpc worker service address should be formatted as host:port."
            )

        self.master_host = master.split(":")[0]
        self.master_port = int(master.split(":")[1])

        if self.config.has_option("worker", "bind_address"):
            self.bind_address = self.config.get("worker", "bind_address")
        else:
            self.bind_address = WorkerConfig.default_bind_address

        if self.config.has_option("worker", "bind_port"):
            self.bind_port = self.config.getint("worker", "bind_port")
        else:
            self.bind_port = WorkerConfig.default_bind_port

        if self.config.has_option("worker", "thrift_threads"):
            self.thrift_threads = self.config.getint("worker", "thrift_threads")
        else:
            self.thrift_threads = WorkerConfig.default_thrift_threads

        if self.config.has_option("worker", "executor_threads"):
            self.executor_threads = self.config.getint("worker", "executor_threads")
        else:
            self.executor_threads = WorkerConfig.default_executor_threads

        if self.config.has_option("worker", "service_address"):
            worker = self.config.get("worker", "worker_service_address")
            if len(worker.split(":")) != 2:
                raise TanitConfigurationException(
                    "Worker rpc service address should be formatted as host:port."
                )

            self.worker_host = master.split(":")[0]
            self.worker_port = int(master.split(":")[1])
        else:
            self.worker_host = socket.gethostname()
            self.worker_port = self.bind_port



#!/usr/bin/env python

import logging as lg
from threading import Thread

from thrift.protocol import TBinaryProtocol
from thrift.server import TServer
from thrift.transport import TSocket
from thrift.transport import TTransport

from ...thrift.worker.service import WorkerService
from ..config.config import WorkerConfig
from .handler import WorkerServiceHandler
from .worker import Worker

_logger = lg.getLogger(__name__)


class WorkerServer(object):
    def __init__(self, config=None):

        self.config = WorkerConfig(config)
        self.config.load()

        self.worker = Worker()
        self.worker.configure(self.config)

        self.stopped = False

    def stop(self):
        self.stopped = True

    def _run(self):
        # Create Service handler
        handler = WorkerServiceHandler(self.worker)

        server = TServer.TThreadedServer(
            WorkerService.Processor(handler),
            TSocket.TServerSocket(self.config.bind_address, self.config.bind_port),
            TTransport.TBufferedTransportFactory(),
            TBinaryProtocol.TBinaryProtocolFactory(),
            daemon=True,
        )

        # Start Tanit server
        server.serve()

    def start(self):

        self.stopped = False

        _logger.info("Stating Tanit worker server.")
        self.daemon = Thread(target=self._run, args=())
        self.daemon.setDaemon(True)
        self.daemon.start()
        _logger.info(
            "Tanit worker server started, listening  at %s:%s",
            self.config.bind_address,
            self.config.bind_port,
        )

        # Start worker services
        try:
            self.worker.start()
        except Exception:
            _logger.exception("Failed to start Tanit worker services.")
            exit(1)

        try:
            while self.daemon.is_alive():
                # Try to join the child thread back to parent for 0.5 seconds
                self.daemon.join(0.5)

                if self.stopped:
                    _logger.info("Tanit worker server stopped, exiting.")
                    break
        except (KeyboardInterrupt, SystemExit):
            _logger.info("Received KeyboardInterrupt Signal.")
        except Exception as e:
            _logger.exception("Fatal server exception : %s, exiting", e)
        finally:
            _logger.info("Stopping Tanit worker server.")
            self.worker.stop()
            _logger.info("Tanit worker server stopped.")
